# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any, Dict, Generator, List, Type, Tuple

from ..core import Serializer, Placeholder, buffered
from .field import Field, OneOfField


class SerializableMeta(type):
    def __new__(mcs,
                name: str,
                bases: Tuple[Type],
                properties: Dict):
        new_properties = dict()
        for base in bases:
            if hasattr(base, '_FIELDS'):
                new_properties.update(base._FIELDS)
        new_properties.update(properties)
        properties = new_properties

        property_to_fields = dict()
        # filter out all fields
        for k, v in properties.items():
            if not isinstance(v, Field):
                continue

            property_to_fields[k] = v
            properties[k] = v
            v._attr_name = k

        properties['_FIELDS'] = property_to_fields
        slots = set(properties.pop('__slots__', set()))
        if property_to_fields:
            slots.add('_FIELD_VALUES')
        properties['__slots__'] = tuple(slots)

        clz = type.__new__(mcs, name, bases, properties)
        return clz


class Serializable(metaclass=SerializableMeta):
    __slots__ = ()

    _FIELDS: Dict[str, Field]
    _FIELD_VALUES: Dict[str, Any]

    def __init__(self, *args, **kwargs):
        setattr(self, '_FIELD_VALUES', dict())

        for slot, arg in zip(self.__slots__, args):  # pragma: no cover
            object.__setattr__(self, slot, arg)

        for key, val in kwargs.items():
            object.__setattr__(self, key, val)

    def __repr__(self):
        values = ', '.join(['{}={!r}'.format(slot, getattr(self, slot, None)) for slot in self.__slots__])
        return '{}({})'.format(self.__class__.__name__, values)


class SerializableSerializer(Serializer):
    """
    Leverage DictSerializer to perform serde.
    """
    serializer_name = 'serializable'

    @classmethod
    def _get_tag_to_values(cls, obj: Serializable):
        fields = obj._FIELDS
        tag_to_values = obj._FIELD_VALUES.copy()

        for field in fields.values():
            tag = field.tag
            try:
                value = tag_to_values[tag]
            except KeyError:
                continue
            if field.on_serialize:
                tag_to_values[tag] = field.on_serialize(value)

        return tag_to_values

    @buffered
    def serialize(self, obj: Serializable, context: Dict):
        tag_to_values = self._get_tag_to_values(obj)

        keys = [None] * len(tag_to_values)
        value_headers = [None] * len(tag_to_values)
        value_sizes = [0] * len(tag_to_values)
        value_buffers = []
        for idx, (key, val) in enumerate(tag_to_values.items()):
            keys[idx] = key
            value_headers[idx], val_buf = yield val
            value_sizes[idx] = len(val_buf)
            value_buffers.extend(val_buf)

        header = {
            'keys': keys,
            'value_headers': value_headers,
            'value_sizes': value_sizes,
            'class': type(obj),
        }
        return header, value_buffers

    def deserialize(self, header: Dict, buffers: List,
                    context: Dict) -> Generator[Any, Any, Serializable]:
        obj_class: Type[Serializable] = header.pop('class')

        tag_to_values = dict()
        pos = 0
        for key, value_header, value_size in \
                zip(header['keys'], header['value_headers'], header['value_sizes']):
            tag_to_values[key] = yield value_header, buffers[pos:pos + value_size]  # noqa: E999
            pos += value_size

        for field in obj_class._FIELDS.values():
            if not isinstance(field, OneOfField):
                try:
                    value = tag_to_values[field.tag]
                except KeyError:
                    continue
                if value is not None:
                    if field.on_deserialize:
                        def cb(v, field_):
                            tag_to_values[field_.tag] = field_.on_deserialize(v)
                    else:
                        def cb(v, field_):
                            tag_to_values[field_.tag] = v

                    if isinstance(value, Placeholder):
                        value.callbacks.append(partial(cb, field_=field))
                    else:
                        cb(value, field)

        obj = obj_class()
        obj._FIELD_VALUES = tag_to_values
        return obj


SerializableSerializer.register(Serializable)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging
import os

from tornado import web

from ... import oscar as mo
from ...utils import get_next_port
from ..core import AbstractService

logger = logging.getLogger(__name__)


class WebActor(mo.Actor):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._web_server = None
        self._web_app = None

        extra_mod_names = self._config.get('extra_discovery_modules') or []
        web_handlers = self._config.get('web_handlers', {})
        for mod_name in extra_mod_names:
            try:
                web_mod = importlib.import_module(mod_name)
                web_handlers.update(getattr(web_mod, 'web_handlers', {}))
            except ImportError:  # pragma: no cover
                pass

    async def __post_create__(self):
        from .indexhandler import handlers as web_handlers

        static_path = os.path.join(os.path.dirname(__file__), 'static')
        supervisor_addr = self.address

        host = self._config.get('host') or '0.0.0.0'
        port = self._config.get('port') or get_next_port()
        self._web_address = f'http://{host}:{port}'
        web_handlers.update(self._config.get('web_handlers', {}))

        handler_kwargs = {'supervisor_addr': supervisor_addr}
        handlers = [
            (r'[^\?\&]*/static/(.*)', web.StaticFileHandler, {'path': static_path})
        ]
        for p, h in web_handlers.items():
            handlers.append((p, h, handler_kwargs))

        retrial = 5
        while retrial:
            try:
                if port is None:
                    port = get_next_port()

                self._web_app = web.Application(handlers)
                self._web_server = self._web_app.listen(port, host)
                logger.info('Mars Web started at %s:%d', host, port)
                break
            except OSError:  # pragma: no cover
                if port is not None:
                    raise
                retrial -= 1
                if retrial == 0:
                    raise

    async def __pre_destroy__(self):
        if self._web_server is not None:
            self._web_server.stop()

    def get_web_address(self):
        web_address = self._web_address
        if os.name == 'nt':
            web_address = web_address.replace('0.0.0.0', '127.0.0.1')
        return web_address


class WebSupervisorService(AbstractService):
    """
    Web service on supervisor.

    Service Configuration
    ---------------------
    {
        "web": {
            "host": "<web host>",
            "port": "<web port>",
            "web_handlers": [
                <web_handlers>,
            ],
            "extra_discovery_modules": [
                "path.to.modules",
            ]
        }
    }
    """
    async def start(self):
        await mo.create_actor(WebActor, config=self._config.get('web', {}),
                              uid=WebActor.default_uid(), address=self._address)

    async def stop(self):
        await mo.destroy_actor(mo.create_actor_ref(
            uid=WebActor.default_uid(), address=self._address))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from .base import AioBase, delegate_to_executor, \
    proxy_method_directly, proxy_property_directly


@delegate_to_executor(
    "close",
    "flush",
    "isatty",
    "read",
    "read1",
    "readinto",
    "readline",
    "readlines",
    "seek",
    "seekable",
    "tell",
    "truncate",
    "writable",
    "write",
    "writelines",
)
@proxy_method_directly("fileno", "readable")
@proxy_property_directly("closed", "name", "mode")
class AioFileObject(AioBase):
    def __aiter__(self):
        return self

    async def __anext__(self):
        """Simulate normal file iteration."""
        line = await self.readline()
        if line:
            return line
        else:
            raise StopAsyncIteration

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        self._file = None


@delegate_to_executor(
    "cat",
    "ls",
    "delete",
    "disk_usage",
    "stat",
    "rm",
    "mv",
    "rename",
    "mkdir",
    "exists",
    "isdir",
    "isfile",
    "read_parquet",
    "walk",
)
@proxy_property_directly("pathsep")
class AioFilesystem(AioBase):
    async def open(self, *args, **kwargs):
        func = functools.partial(self._file.open, *args, **kwargs)
        file = await self._loop.run_in_executor(self._executor, func)
        return AioFileObject(file)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ....utils import calc_size_by_str
from ...core import AbstractService
from .workerslot import WorkerSlotManagerActor
from .quota import WorkerQuotaManagerActor
from .execution import SubtaskExecutionActor, DEFAULT_SUBTASK_MAX_RETRIES


class SchedulingWorkerService(AbstractService):
    """
    Scheduling service on worker.

    Service Configuration
    ---------------------
    {
        "scheduling": {
            "mem_quota_size": "80%",
            "mem_hard_limit": "95%",
            "enable_kill_slot": true,
            "subtask_max_retries": 1
        }
    }
    """
    async def start(self):
        from .... import resource as mars_resource
        scheduling_config = self._config.get('scheduling', {})
        address = self._address

        total_mem = mars_resource.virtual_memory().total
        mem_quota_size = calc_size_by_str(
            scheduling_config.get('mem_quota_size', '80%'), total_mem)
        mem_hard_limit = calc_size_by_str(
            scheduling_config.get('mem_hard_limit', '95%'), total_mem)
        enable_kill_slot = scheduling_config.get('enable_kill_slot', True)
        subtask_max_retries = scheduling_config.get('subtask_max_retries', DEFAULT_SUBTASK_MAX_RETRIES)

        await mo.create_actor(WorkerSlotManagerActor,
                              uid=WorkerSlotManagerActor.default_uid(),
                              address=address)
        await mo.create_actor(WorkerQuotaManagerActor,
                              default_config=dict(quota_size=mem_quota_size,
                                                  hard_limit=mem_hard_limit,
                                                  enable_kill_slot=enable_kill_slot),
                              uid=WorkerQuotaManagerActor.default_uid(),
                              address=address)
        await mo.create_actor(SubtaskExecutionActor,
                              subtask_max_retries=subtask_max_retries,
                              enable_kill_slot=enable_kill_slot,
                              uid=SubtaskExecutionActor.default_uid(),
                              address=address)

    async def stop(self):
        address = self._address

        await mo.destroy_actor(mo.create_actor_ref(
            uid=SubtaskExecutionActor.default_uid(), address=address))
        await mo.destroy_actor(mo.create_actor_ref(
            uid=WorkerQuotaManagerActor.default_uid(), address=address))
        await mo.destroy_actor(mo.create_actor_ref(
            uid=WorkerSlotManagerActor.default_uid(), address=address))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral

import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import KeyField, TupleField, AnyField
from ...tensor import tensor as astensor
from ...utils import has_unknown_shape
from ..core import TENSOR_TYPE
from ..operands import TensorHasInput, TensorOperandMixin
from ..utils import filter_inputs
from .core import process_index


class TensorIndexSetValue(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.INDEXSETVALUE

    _input = KeyField('input')
    _indexes = TupleField('indexes')
    _value = AnyField('value')

    def __init__(self, indexes=None, value=None, **kw):
        super().__init__(_indexes=indexes, _value=value, **kw)

    @property
    def indexes(self):
        return self._indexes

    @property
    def value(self):
        return self._value

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs[1:])
        new_indexes = [next(inputs_iter) if isinstance(index, ENTITY_TYPE) else index
                       for index in self._indexes]
        self._indexes = tuple(new_indexes)
        if isinstance(self._value, ENTITY_TYPE):
            self._value = next(inputs_iter)

    def __call__(self, a, index, value):
        inputs = filter_inputs([a] + list(index) + [value])
        self._indexes = tuple(index)
        self._value = value
        return self.new_tensor(inputs, a.shape, order=a.order)

    def on_output_modify(self, new_output):
        return new_output

    def on_input_modify(self, new_input):
        new_op = self.copy().reset_key()
        new_inputs = [new_input] + self.inputs[1:]
        return new_op.new_tensor(new_inputs, shape=self.outputs[0].shape)

    @classmethod
    def tile(cls, op: "TensorIndexSetValue"):
        from ..base import broadcast_to
        from .getitem import _getitem_nocheck

        tensor = op.outputs[0]
        value = op.value
        indexed = yield from recursive_tile(
            _getitem_nocheck(op.input, op.indexes, convert_bool_to_fancy=False))
        is_value_tensor = isinstance(value, TENSOR_TYPE)

        if is_value_tensor and value.ndim > 0:
            if has_unknown_shape(indexed, value):
                yield indexed.chunks + [indexed]

            value = yield from recursive_tile(
                broadcast_to(value, indexed.shape).astype(op.input.dtype, copy=False))
            nsplits = indexed.nsplits
            value = yield from recursive_tile(value.rechunk(nsplits))

        chunk_mapping = {c.op.input.index: c for c in indexed.chunks}
        out_chunks = []
        for chunk in indexed.op.input.chunks:
            index_chunk = chunk_mapping.get(chunk.index)
            if index_chunk is None:
                out_chunks.append(chunk)
                continue

            if is_value_tensor:
                if value.ndim > 0:
                    value_chunk = value.cix[index_chunk.index]
                else:
                    value_chunk = value.chunks[0]
            else:
                # non tensor
                value_chunk = value
            chunk_op = TensorIndexSetValue(dtype=op.dtype, sparse=op.sparse,
                                           indexes=tuple(index_chunk.op.indexes),
                                           value=value_chunk)
            chunk_inputs = filter_inputs([chunk] + index_chunk.op.indexes + [value_chunk])
            out_chunk = chunk_op.new_chunk(chunk_inputs, shape=chunk.shape,
                                           index=chunk.index, order=tensor.order)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape, order=tensor.order,
                                  chunks=out_chunks, nsplits=op.input.nsplits)

    @classmethod
    def execute(cls, ctx, op):
        indexes = [ctx[index.key] if hasattr(index, 'key') else index
                   for index in op.indexes]
        input_ = ctx[op.inputs[0].key].copy()
        value = ctx[op.value.key] if hasattr(op.value, 'key') else op.value
        if hasattr(input_, 'flags') and not input_.flags.writeable:
            input_.setflags(write=True)
        input_[tuple(indexes)] = value
        ctx[op.outputs[0].key] = input_


def _check_support(index):
    if isinstance(index, (slice, Integral)):
        pass
    elif isinstance(index, (np.ndarray, TENSOR_TYPE)) and index.dtype == np.bool_:
        pass
    else:  # pragma: no cover
        raise NotImplementedError('Only slice, int, or bool indexing '
                                  f'supported by now, got {type(index)}')


def _setitem(a, item, value):
    index = process_index(a.ndim, item, convert_bool_to_fancy=False)
    if not (np.isscalar(value) or (isinstance(value, tuple) and a.dtype.fields)):
        # do not convert for tuple when dtype is record type.
        value = astensor(value)

    for ix in index:
        _check_support(ix)

    # __setitem__ on a view should be still a view, see GH #732.
    op = TensorIndexSetValue(dtype=a.dtype, sparse=a.issparse(),
                             indexes=tuple(index), value=value,
                             create_view=a.op.create_view)
    ret = op(a, index, value)
    a.data = ret.data



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import inspect
import logging
import sys
cimport cython
from typing import AsyncGenerator

from .errors import Return
from .utils cimport is_async_generator
from .utils import create_actor_ref


CALL_METHOD_DEFAULT = 0
CALL_METHOD_BATCH = 1

logger = logging.getLogger(__name__)

cdef:
    bint _log_unhandled_errors = False
    bint _log_cycle_send = False


def set_debug_options(options):
    global _log_unhandled_errors, _log_cycle_send
    if options is None:
        _log_unhandled_errors = _log_cycle_send = False
    else:
        _log_unhandled_errors = options.log_unhandled_errors
        _log_cycle_send = options.log_cycle_send


cdef class ActorRef:
    """
    Reference of an Actor at user side
    """
    def __init__(self, str address, object uid):
        self.uid = uid
        self.address = address
        self._methods = dict()

    cdef __send__(self, object message):
        from .context import get_context
        ctx = get_context()
        return ctx.send(self, message)

    cdef __tell__(self, object message, object delay=None):
        from .context import get_context
        ctx = get_context()
        return ctx.send(self, message, wait_response=False)

    def destroy(self, object callback=None):
        from .context import get_context
        ctx = get_context()
        return ctx.destroy_actor(self)

    def __getstate__(self):
        return self.address, self.uid

    def __setstate__(self, state):
        self.address, self.uid = state

    def __reduce__(self):
        return self.__class__, self.__getstate__()

    def __getattr__(self, str item):
        if item.startswith('_'):
            return object.__getattribute__(self, item)

        try:
            return self._methods[item]
        except KeyError:
            method = self._methods[item] = ActorRefMethod(self, item)
            return method

    def __hash__(self):
        return hash((type(self), self.address, self.uid))

    def __eq__(self, other):
        return isinstance(other, ActorRef) and \
               other.address == self.address and \
               other.uid == self.uid

    def __repr__(self):
        return 'ActorRef(uid={!r}, address={!r})'.format(self.uid, self.address)


cdef class _DelayedArgument:
    cdef readonly tuple arguments

    def __init__(self, tuple arguments):
        self.arguments = arguments


cdef class ActorRefMethod:
    """
    Wrapper for an Actor method at client
    """
    cdef ActorRef ref
    cdef object method_name

    def __init__(self, ref, method_name):
        self.ref = ref
        self.method_name = method_name

    def __call__(self, *args, **kwargs):
        return self.send(*args, **kwargs)

    def send(self, *args, **kwargs):
        arg_tuple = (self.method_name, CALL_METHOD_DEFAULT, args, kwargs)
        return self.ref.__send__(arg_tuple)

    def tell(self, *args, **kwargs):
        arg_tuple = (self.method_name, CALL_METHOD_DEFAULT, args, kwargs)
        return self.ref.__tell__(arg_tuple)

    def delay(self, *args, **kwargs):
        arg_tuple = (self.method_name, CALL_METHOD_DEFAULT, args, kwargs)
        return _DelayedArgument(arg_tuple)

    def batch(self, *delays, send=True):
        cdef:
            list args_list = list()
            list kwargs_list = list()

        last_method = None
        for delay in delays:
            method, _call_method, args, kwargs = delay.arguments
            if last_method is not None and method != last_method:
                raise ValueError('Does not support calling multiple methods in batch')
            last_method = method

            args_list.append(args)
            kwargs_list.append(kwargs)
        if last_method is None:
            last_method = self.method_name

        if send:
            return self.ref.__send__((last_method, CALL_METHOD_BATCH,
                                      (args_list, kwargs_list), {}))
        else:
            return self.ref.__tell__((last_method, CALL_METHOD_BATCH,
                                      (args_list, kwargs_list), {}))

    def tell_delay(self, *args, delay=None, ignore_conn_fail=True, **kwargs):
        async def delay_fun():
            try:
                await asyncio.sleep(delay)
                await self.ref.__tell__((self.method_name, CALL_METHOD_DEFAULT, args, kwargs))
            except Exception as ex:
                if ignore_conn_fail and isinstance(ex, ConnectionRefusedError):
                    return

                logger.error(f'Error {type(ex)} occurred when calling {self.method_name} '
                             f'on {self.ref.uid} at {self.ref.address} with tell_delay')
                raise

        return asyncio.create_task(delay_fun())


cdef class _BaseActor:
    """
    Base Mars actor class, user methods implemented as methods
    """
    def __cinit__(self, *args, **kwargs):
        self._lock = self._create_lock()

    def _create_lock(self):
        raise NotImplementedError

    @property
    def uid(self):
        return self._uid

    @uid.setter
    def uid(self, uid):
        self._uid = uid
        
    def _set_uid(self, uid):
        self._uid = uid

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, addr):
        self._address = addr

    def _set_address(self, addr):
        self._address = addr

    cpdef ActorRef ref(self):
        return create_actor_ref(self._address, self._uid)

    async def _handle_actor_result(self, result):
        cdef int result_pos
        cdef tuple res_tuple
        cdef list tasks, values
        cdef object coro
        cdef bint extract_tuple = False
        cdef bint cancelled = False
        cdef set dones, pending

        if inspect.isawaitable(result):
            result = await result
        elif is_async_generator(result):
            result = (result,)
            extract_tuple = True

        if type(result) is tuple:
            res_tuple = result
            tasks = []
            values = []
            for res_item in res_tuple:
                if is_async_generator(res_item):
                    value = asyncio.create_task(self._run_actor_async_generator(res_item))
                    tasks.append(value)
                elif inspect.isawaitable(res_item):
                    value = asyncio.create_task(res_item)
                    tasks.append(value)
                else:
                    value = res_item
                values.append(value)

            if len(tasks) > 0:
                try:
                    dones, pending = await asyncio.wait(tasks)
                except asyncio.CancelledError:
                    cancelled = True
                    for task in tasks:
                        task.cancel()
                    # wait till all tasks return cancelled
                    dones, pending = await asyncio.wait(tasks)

                if extract_tuple:
                    result = list(dones)[0].result()
                else:
                    result = tuple(t.result() if t in dones else t for t in values)

                if cancelled:
                    # raise in case no CancelledError raised
                    raise asyncio.CancelledError

        return result

    async def _run_actor_async_generator(self, gen: AsyncGenerator):
        """
        Run an async generator under Actor lock
        """
        cdef tuple res_tuple
        cdef bint is_exception = False
        cdef object res
        cdef object message_trace = None, pop_message_trace = None, set_message_trace = None

        from .debug import pop_message_trace, set_message_trace, debug_async_timeout
        try:
            res = None
            while True:
                with debug_async_timeout('actor_lock_timeout',
                                         'async_generator %r hold lock timeout', gen):
                    async with self._lock:
                        if not is_exception:
                            res = await gen.asend(res)
                        else:
                            res = await gen.athrow(*res)
                try:
                    if _log_cycle_send:
                        message_trace = pop_message_trace()

                    res = await self._handle_actor_result(res)
                    is_exception = False
                except:
                    res = sys.exc_info()
                    is_exception = True
                finally:
                    if _log_cycle_send:
                        set_message_trace(message_trace)
        except Return as ex:
            return ex.value
        except StopAsyncIteration as ex:
            return

    async def __post_create__(self):
        """
        Method called after actor creation
        """
        pass

    async def __pre_destroy__(self):
        """
        Method called before actor destroy
        """
        pass

    async def __on_receive__(self, tuple message):
        """
        Handle message from other actors and dispatch them to user methods

        Parameters
        ----------
        message : tuple
            Message shall be (method_name,) + args + (kwargs,)
        """
        from .debug import debug_async_timeout
        try:
            method, call_method, args, kwargs = message
            if call_method == CALL_METHOD_DEFAULT:
                func = getattr(self, method)
                with debug_async_timeout('actor_lock_timeout',
                                         "Method %s of actor %s hold lock timeout.",
                                         method, self.uid):
                    async with self._lock:
                        result = func(*args, **kwargs)
                        if asyncio.iscoroutine(result):
                            result = await result
            elif call_method == CALL_METHOD_BATCH:
                func = getattr(self, method)
                with debug_async_timeout('actor_lock_timeout',
                                         "Batch method %s of actor %s hold lock timeout, batch size %s.",
                                         method, self.uid, len(args)):
                    async with self._lock:
                        delays = []
                        for s_args, s_kwargs in zip(*args):
                            delays.append(func.delay(*s_args, **s_kwargs))
                        result = func.batch(*delays)
                        if asyncio.iscoroutine(result):
                            result = await result
            else:  # pragma: no cover
                raise ValueError(f'call_method {call_method} not valid')

            return await self._handle_actor_result(result)
        except Exception as ex:
            if _log_unhandled_errors:
                from .debug import logger as debug_logger
                debug_logger.exception('Got unhandled error when handling message %r'
                                       'in actor %s at %s', message, self.uid, self.address)
            raise ex


# The @cython.binding(True) is for ray getting members.
# The value is True by default after cython >= 3.0.0
@cython.binding(True)
cdef class _Actor(_BaseActor):
    def _create_lock(self):
        return asyncio.locks.Lock()


cdef class _FakeLock:
    async def __aenter__(self):
        pass

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


# The @cython.binding(True) is for ray getting members.
# The value is True by default after cython >= 3.0.0
@cython.binding(True)
cdef class _StatelessActor(_BaseActor):
    def _create_lock(self):
        return _FakeLock()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...serialization.serializables import FieldTypes, TupleField
from ...utils import tokenize
from .core import EntityData, Entity


class ChunkData(EntityData):
    __slots__ = ()

    # optional fields
    _index = TupleField('index', FieldTypes.uint32)

    def __repr__(self):
        if self.op.stage is None:
            return f'Chunk <op={type(self.op).__name__}, ' \
                   f'key={self.key}>'
        else:
            return f'Chunk <op={type(self.op).__name__}, ' \
                   f'stage={self.op.stage.name}, key={self.key}>'

    @property
    def index(self):
        return getattr(self, '_index', None)

    @property
    def device(self):
        return self.op.device

    def _update_key(self):
        object.__setattr__(self, '_key', tokenize(
            type(self).__name__, *(getattr(self, k, None) for k in self._keys_
                                   if k != '_index')))


class Chunk(Entity):
    _allow_data_type_ = (ChunkData,)


CHUNK_TYPE = (Chunk, ChunkData)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noinspection PyUnresolvedReferences
from ..typing import ChunkType, TileableType, EntityType, OperandType
from .entity import Entity, EntityData, ENTITY_TYPE, \
    Chunk, ChunkData, CHUNK_TYPE, \
    Tileable, TileableData, TILEABLE_TYPE, \
    Object, ObjectData, ObjectChunk, ObjectChunkData, OBJECT_TYPE, OBJECT_CHUNK_TYPE, \
    FuseChunk, FuseChunkData, FUSE_CHUNK_TYPE, \
    OutputType, register_output_types, get_output_types, \
    register_fetch_class, get_fetch_class, get_tileable_types, get_chunk_types, \
    HasShapeTileable, HasShapeTileableData, ExecutableTuple, _ExecuteAndFetchMixin, \
    NotSupportTile, register, unregister, tile, recursive_tile
# noinspection PyUnresolvedReferences
from .graph import DirectedGraph, DAG, GraphContainsCycleError, \
    TileableGraph, ChunkGraph, TileableGraphBuilder, ChunkGraphBuilder
from .mode import enter_mode, is_build_mode, is_eager_mode, is_kernel_mode



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import importlib
import inspect
from abc import ABC, ABCMeta, abstractmethod
from typing import Any, Callable, Optional, Type, Union

from .field_type import AbstractFieldType, FieldTypes, \
    ListType, TupleType, DictType, ReferenceType

_notset = object()


class Field(ABC):
    __slots__ = '_tag', '_default_value', '_default_factory', \
                '_attr_name', '_on_serialize', '_on_deserialize'

    _tag: str
    _default_value: Any
    _default_factory: Optional[Callable]
    _attr_name: str  # attribute name that set to

    def __init__(self,
                 tag: str,
                 default: Any = _notset,
                 default_factory: Optional[Callable] = None,
                 on_serialize: Callable[[Any], Any] = None,
                 on_deserialize: Callable[[Any], Any] = None,
                 attr_name: str = None):
        if default is not _notset and default_factory is not None:  # pragma: no cover
            raise ValueError('default and default_factory can not be specified both')

        self._tag = tag
        self._default_value = default
        self._default_factory = default_factory
        self._on_serialize = on_serialize
        self._on_deserialize = on_deserialize
        self._attr_name = attr_name

    @property
    def tag(self):
        return self._tag

    @property
    def on_serialize(self):
        return self._on_serialize

    @property
    def on_deserialize(self):
        return self._on_deserialize

    @property
    def attr_name(self):
        return self._attr_name

    @property
    @abstractmethod
    def field_type(self) -> AbstractFieldType:
        """
        Field type.

        Returns
        -------
        field_type : AbstractFieldType
             Field type.
        """

    def __get__(self, instance, owner):
        tag_to_values = instance._FIELD_VALUES
        try:
            return tag_to_values[self._tag]
        except KeyError:
            if self._default_value is not _notset:
                return self._default_value
            elif self._default_factory is not None:
                val = tag_to_values[self._tag] = self._default_factory()
                return val
            else:
                raise AttributeError(
                    f"'{type(instance)}' has no attribute {self._attr_name}")

    def __set__(self, instance, value):
        from ...core import is_kernel_mode

        if not is_kernel_mode():
            field_type = self.field_type
            try:
                to_check_value = value
                if to_check_value is not None and self._on_serialize:
                    to_check_value = self._on_serialize(to_check_value)
                field_type.validate(to_check_value)
            except (TypeError, ValueError) as e:
                raise type(e)(f'Failed to set `{self._attr_name}`: {str(e)}')
        instance._FIELD_VALUES[self._tag] = value

    def __delete__(self, instance):
        del instance._FIELD_VALUES[self._tag]


class AnyField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.any


class IdentityField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.string


class BoolField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.bool


class Int8Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int8


class Int16Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int16


class Int32Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int32


class Int64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.int64


class UInt8Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint8


class UInt16Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint16


class UInt32Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint32


class UInt64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.uint64


class Float16Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.float16


class Float32Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.float32


class Float64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.float64


class Complex64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.complex64


class Complex128Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.complex128


class StringField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.string


class BytesField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.bytes


class KeyField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.key


class NDArrayField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.ndarray


class Datetime64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.datetime


class Timedelta64Field(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.timedelta


class DataTypeField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.dtype


class IndexField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.index


class SeriesField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.series


class DataFrameField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.dataframe


class SliceField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.slice


class FunctionField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.function


class NamedTupleField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.namedtuple


class TZInfoField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.tzinfo


class IntervalArrayField(Field):
    __slots__ = ()

    @property
    def field_type(self) -> AbstractFieldType:
        return FieldTypes.interval_array


class _CollectionField(Field, metaclass=ABCMeta):
    __slots__ = '_field_type',

    def __init__(self,
                 tag: str,
                 field_type: AbstractFieldType = None,
                 default: Any = _notset,
                 default_factory: Optional[Callable] = None,
                 on_serialize: Callable[[Any], Any] = None,
                 on_deserialize: Callable[[Any], Any] = None):
        super().__init__(tag, default=default, default_factory=default_factory,
                         on_serialize=on_serialize, on_deserialize=on_deserialize)
        if field_type is None:
            field_type = FieldTypes.any
        if not isinstance(field_type, ListType):
            field_type = self._collection_type()(field_type, ...)
        self._field_type = field_type

    @classmethod
    @abstractmethod
    def _collection_type(cls) -> AbstractFieldType:
        """
        Collection type.

        Returns
        -------
        collection_type
        """

    @property
    def field_type(self) -> Type[AbstractFieldType]:
        return self._field_type


class ListField(_CollectionField):
    __slots__ = ()

    @classmethod
    def _collection_type(cls) -> Type[AbstractFieldType]:
        return ListType


class TupleField(_CollectionField):
    __slots__ = ()

    @classmethod
    def _collection_type(cls) -> Type[AbstractFieldType]:
        return TupleType


class DictField(Field):
    __slots__ = '_field_type',

    def __init__(self,
                 tag: str,
                 key_type: AbstractFieldType = None,
                 value_type: AbstractFieldType = None,
                 default: Any = _notset,
                 default_factory: Optional[Callable] = None,
                 on_serialize: Callable[[Any], Any] = None,
                 on_deserialize: Callable[[Any], Any] = None):
        super().__init__(tag, default=default, default_factory=default_factory,
                         on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._field_type = DictType(key_type, value_type)

    @property
    def field_type(self) -> AbstractFieldType:
        return self._field_type


class ReferenceField(Field):
    __slots__ = '_reference_type', '_field_type'

    def __init__(self,
                 tag: str,
                 reference_type: Union[str, Type] = None,
                 default: Any = _notset,
                 on_serialize: Callable[[Any], Any] = None,
                 on_deserialize: Callable[[Any], Any] = None):
        super().__init__(tag, default=default, on_serialize=on_serialize,
                         on_deserialize=on_deserialize)
        self._reference_type = reference_type

        if not isinstance(reference_type, str):
            self._field_type = ReferenceType(reference_type)
        else:
            # need to bind dynamically
            self._field_type = None

    @property
    def field_type(self) -> AbstractFieldType:
        return self._field_type

    def get_field_type(self, instance):
        if self._field_type is None:
            # bind dynamically
            if self._reference_type == 'self':
                reference_type = type(instance)
            elif isinstance(self._reference_type, str) and \
                    '.' in self._reference_type:
                module, name = self._reference_type.rsplit('.', 1)
                reference_type = getattr(importlib.import_module(module), name)
            else:
                module = inspect.getmodule(instance)
                reference_type = getattr(module, self._reference_type)
            self._field_type = ReferenceType(reference_type)
        return self._field_type

    def __set__(self, instance, value):
        from ...core import is_kernel_mode

        if self._field_type is None:
            if not is_kernel_mode():
                field_type = self.get_field_type(instance)
                try:
                    to_check_value = value
                    if to_check_value is not None and self._on_serialize:
                        to_check_value = self._on_serialize(to_check_value)
                    field_type.validate(to_check_value)
                except (TypeError, ValueError) as e:
                    if not self._attr_name:
                        raise
                    else:
                        raise type(e)(f'Failed to set `{self._attr_name}`: {str(e)}')
            instance._FIELD_VALUES[self._tag] = value
        else:
            super().__set__(instance, value)


class OneOfField(Field):
    __slots__ = '_reference_fields'

    def __init__(self,
                 tag: str,
                 default: Any = _notset,
                 on_serialize: Callable[[Any], Any] = None,
                 on_deserialize: Callable[[Any], Any] = None,
                 attr_name: str = None,
                 **tag_to_reference_types):
        super().__init__(
            tag, default=default, on_serialize=on_serialize,
            on_deserialize=on_deserialize,
            attr_name=attr_name)
        self._reference_fields = [
            ReferenceField(t, ref_type) for t, ref_type
            in tag_to_reference_types.items()]

    @property
    def reference_fields(self):
        return self._reference_fields

    @property
    def field_type(self) -> AbstractFieldType:  # pragma: no cover
        # takes no effect here, just return AnyType()
        # we will do check in __set__ instead
        return FieldTypes.any

    def __set__(self, instance, value):
        field_values = instance._FIELD_VALUES
        for reference_field in self._reference_fields:
            try:
                to_check_value = value
                if to_check_value is not None and self._on_serialize:
                    to_check_value = self._on_serialize(to_check_value)
                reference_field.get_field_type(instance).validate(to_check_value)
                field_values[reference_field.tag] = value
                return
            except TypeError:
                continue
        valid_types = list(itertools.chain(*[r.get_field_type(instance).valid_types
                                             for r in self._reference_fields]))
        raise TypeError(f'Failed to set `{self._attr_name}`: type of instance cannot '
                        f'match any of {valid_types}, got {type(value)}')

    def __get__(self, instance, owner):
        field_values = instance._FIELD_VALUES
        for reference_field in self._reference_fields:
            try:
                return field_values[reference_field.tag]
            except KeyError:
                continue
        raise AttributeError('cannot get attribute, '
                             'maybe value not set before')

    def __delete__(self, instance):
        field_values = instance._FIELD_VALUES
        for reference_field in self._reference_fields:
            try:
                del field_values[reference_field.tag]
                return
            except KeyError:
                continue
        raise AttributeError('cannot delete attribute, '
                             'maybe value not set before')



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import sys
from collections import defaultdict
from typing import Any, Dict, List, Optional, Type

from .... import oscar as mo
from ....core import ChunkGraph, OperandType, enter_mode
from ....core.context import get_context, set_context
from ....core.operand import Fetch, FetchShuffle, \
    MapReduceOperand, VirtualOperand, execute
from ....optimization.physical import optimize
from ....typing import BandType
from ....utils import get_chunk_key_to_data_keys
from ...context import ThreadedServiceContext
from ...meta.api import MetaAPI
from ...storage import StorageAPI
from ...session import SessionAPI
from ...task import TaskAPI, task_options
from ..core import Subtask, SubtaskStatus, SubtaskResult

logger = logging.getLogger(__name__)


class DataStore(dict):
    def __getattr__(self, attr):
        ctx = get_context()
        return getattr(ctx, attr)


class SubtaskProcessor:
    _chunk_graph: ChunkGraph
    _chunk_key_to_data_keys: Dict[str, List[str]]

    def __init__(self,
                 subtask: Subtask,
                 session_api: SessionAPI,
                 storage_api: StorageAPI,
                 meta_api: MetaAPI,
                 band: BandType,
                 supervisor_address: str,
                 engines: List[str] = None):
        self.subtask = subtask
        self._session_id = self.subtask.session_id
        self._chunk_graph = subtask.chunk_graph
        self._actual_chunk_count = len([
            chunk for chunk in subtask.chunk_graph
            if not isinstance(chunk.op, (Fetch, FetchShuffle))
        ])
        self._band = band
        self._supervisor_address = supervisor_address
        self._engines = engines if engines is not None else \
            task_options.runtime_engines

        # result
        self.result = SubtaskResult(
            subtask_id=subtask.subtask_id,
            session_id=subtask.session_id,
            task_id=subtask.task_id,
            status=SubtaskStatus.pending,
            bands=[self._band],
            progress=0.0)
        self.is_done = asyncio.Event()

        # status and intermediate states
        # operand progress, from op key to progress
        self._op_progress: Dict[str, float] = defaultdict(lambda: 0.0)
        # temp data store that holds chunk data during computation
        self._datastore = DataStore()
        # chunk key to real data keys
        self._chunk_key_to_data_keys = dict()

        # other service APIs
        self._session_api = session_api
        self._storage_api = storage_api
        self._meta_api = meta_api

    @property
    def status(self):
        return self.result.status

    @property
    def subtask_id(self):
        return self.subtask.subtask_id

    async def _load_input_data(self):
        keys, gets, accept_nones = [], [], []
        for chunk in self._chunk_graph.iter_indep():
            if isinstance(chunk.op, Fetch):
                keys.append(chunk.key)
                gets.append(self._storage_api.get.delay(chunk.key))
                accept_nones.append(True)
            elif isinstance(chunk.op, FetchShuffle):
                for key in self._chunk_key_to_data_keys[chunk.key]:
                    keys.append(key)
                    gets.append(self._storage_api.get.delay(key, error='ignore'))
                    accept_nones.append(False)
        if keys:
            logger.debug('Start getting input data, keys: %s, '
                         'subtask id: %s', keys, self.subtask.subtask_id)
            inputs = await self._storage_api.get.batch(*gets)
            self._datastore.update(
                {key: get for key, get, accept_none in zip(keys, inputs, accept_nones)
                 if accept_none or get is not None}
            )
            logger.debug('Finish getting input data keys: %s, '
                         'subtask id: %s', keys, self.subtask.subtask_id)
        return keys

    @staticmethod
    async def notify_task_manager_result(supervisor_address: str,
                                         result: SubtaskResult):
        task_api = await TaskAPI.create(
            result.session_id, supervisor_address)
        # notify task service
        await task_api.set_subtask_result(result)

    def _init_ref_counts(self) -> Dict[str, int]:
        chunk_graph = self._chunk_graph
        ref_counts = defaultdict(lambda: 0)
        # set 1 for result chunks
        for result_chunk in chunk_graph.result_chunks:
            ref_counts[result_chunk.key] += 1
        # iter graph to set ref counts
        for chunk in chunk_graph:
            ref_counts[chunk.key] += chunk_graph.count_successors(chunk)
        return ref_counts

    async def _async_execute_operand(self,
                                     ctx: Dict[str, Any],
                                     op: OperandType):
        if not isinstance(op, (Fetch, FetchShuffle)):
            self._op_progress[op.key] = 0.0
        get_context().set_running_operand_key(self._session_id, op.key)
        return asyncio.to_thread(self._execute_operand, ctx, op)

    def set_op_progress(self, op_key: str, progress: float):
        if op_key in self._op_progress:  # pragma: no branch
            self._op_progress[op_key] = progress

    @enter_mode(build=False, kernel=True)
    def _execute_operand(self,
                         ctx: Dict[str, Any],
                         op: OperandType):  # noqa: R0201  # pylint: disable=no-self-use
        return execute(ctx, op)

    async def _execute_graph(self, chunk_graph: ChunkGraph):
        loop = asyncio.get_running_loop()
        ref_counts = self._init_ref_counts()

        # from data_key to results
        for chunk in chunk_graph.topological_iter():
            if chunk.key not in self._datastore:
                # since `op.execute` may be a time-consuming operation,
                # we make it run in a thread pool to not block current thread.
                logger.debug('Start executing operand: %s,'
                             'chunk: %s, subtask id: %s', chunk.op, chunk,
                             self.subtask.subtask_id)
                future = asyncio.create_task(
                    await self._async_execute_operand(self._datastore, chunk.op))
                to_wait = loop.create_future()

                def cb(fut):
                    if not to_wait.done():
                        if fut.exception():
                            to_wait.set_exception(fut.exception())
                        else:
                            to_wait.set_result(fut.result())
                future.add_done_callback(cb)

                try:
                    await to_wait
                    logger.debug('Finish executing operand: %s,'
                                 'chunk: %s, subtask id: %s', chunk.op, chunk,
                                 self.subtask.subtask_id)
                except asyncio.CancelledError:
                    logger.debug('Receive cancel instruction for operand: %s,'
                                 'chunk: %s, subtask id: %s', chunk.op, chunk,
                                 self.subtask.subtask_id)
                    # wait for this computation to finish
                    await future
                    # if cancelled, stop next computation
                    logger.debug('Cancelled operand: %s, chunk: %s, '
                                 'subtask id: %s', chunk.op, chunk,
                                 self.subtask.subtask_id)
                    self.result.status = SubtaskStatus.cancelled
                    raise

            self.set_op_progress(chunk.op.key, 1.0)

            for inp in chunk.inputs:
                ref_counts[inp.key] -= 1
                if ref_counts[inp.key] == 0:
                    # ref count reaches 0, remove it
                    for key in self._chunk_key_to_data_keys[inp.key]:
                        del self._datastore[key]

    async def _unpin_data(self, data_keys):
        # unpin input keys
        unpins = []
        for key in data_keys:
            if isinstance(key, tuple):
                # a tuple key means it's a shuffle key,
                # some shuffle data is None and not stored in storage
                unpins.append(self._storage_api.unpin.delay(key, error='ignore'))
            else:
                unpins.append(self._storage_api.unpin.delay(key))
        await self._storage_api.unpin.batch(*unpins)

    async def _store_data(self, chunk_graph: ChunkGraph):
        # skip virtual operands for result chunks
        result_chunks = [c for c in chunk_graph.result_chunks
                         if not isinstance(c.op, VirtualOperand)]

        # store data into storage
        puts = []
        stored_keys = []
        for result_chunk in result_chunks:
            data_key = result_chunk.key
            if data_key in self._datastore:
                # non shuffle op
                stored_keys.append(data_key)
                result_data = self._datastore[data_key]
                # update meta
                if not isinstance(result_data, tuple):
                    result_chunk.params = result_chunk.get_params_from_data(result_data)

                put = self._storage_api.put.delay(data_key, result_data)
                puts.append(put)
            else:
                assert isinstance(result_chunk.op, MapReduceOperand)
                keys = [store_key for store_key in self._datastore
                        if isinstance(store_key, tuple) and store_key[0] == data_key]
                for key in keys:
                    stored_keys.append(key)
                    result_data = self._datastore[key]
                    put = self._storage_api.put.delay(key, result_data)
                    puts.append(put)
        logger.debug('Start putting data keys: %s, '
                     'subtask id: %s', stored_keys, self.subtask.subtask_id)
        data_key_to_store_size = dict()
        data_key_to_memory_size = dict()
        data_key_to_object_id = dict()
        if puts:
            put_infos = asyncio.create_task(self._storage_api.put.batch(*puts))
            try:
                store_infos = await put_infos
                for store_key, store_info in zip(stored_keys, store_infos):
                    data_key_to_store_size[store_key] = store_info.store_size
                    data_key_to_memory_size[store_key] = store_info.memory_size
                    data_key_to_object_id[store_key] = store_info.object_id
                logger.debug('Finish putting data keys: %s, '
                             'subtask id: %s', stored_keys, self.subtask.subtask_id)
            except asyncio.CancelledError:
                logger.debug('Cancelling put data keys: %s, '
                             'subtask id: %s', stored_keys, self.subtask.subtask_id)
                put_infos.cancel()

                logger.debug('Cancelled put data keys: %s, '
                             'subtask id: %s', stored_keys, self.subtask.subtask_id)
                self.result.status = SubtaskStatus.cancelled
                raise

        # clear data
        self._datastore = dict()
        return stored_keys, data_key_to_store_size, data_key_to_memory_size, data_key_to_object_id

    async def _store_meta(self,
                          chunk_graph: ChunkGraph,
                          stored_keys: List,
                          data_key_to_store_size: Dict,
                          data_key_to_memory_size: Dict,
                          data_key_to_object_id: Dict):
        key_to_result_chunk = {c.key: c for c in chunk_graph.result_chunks}
        # store meta
        set_chunk_metas = []
        result_data_size = 0
        for chunk_key in stored_keys:
            if isinstance(chunk_key, tuple):
                result_chunk = key_to_result_chunk[chunk_key[0]]
            else:
                result_chunk = key_to_result_chunk[chunk_key]
            store_size = data_key_to_store_size[chunk_key]
            memory_size = data_key_to_memory_size[chunk_key]
            result_data_size += memory_size
            object_ref = data_key_to_object_id[chunk_key]
            set_chunk_metas.append(
                self._meta_api.set_chunk_meta.delay(
                    result_chunk, memory_size=memory_size,
                    store_size=store_size, bands=[self._band],
                    chunk_key=chunk_key, object_ref=object_ref))
        for chunk in chunk_graph.result_chunks:
            if chunk.key not in data_key_to_store_size:
                # mapper, set meta, so that storage can make sure
                # this operand is executed, some sub key is absent
                # due to it's empty actually
                set_chunk_metas.append(self._meta_api.set_chunk_meta.delay(
                    chunk, memory_size=0, store_size=0, bands=[self._band]))
        logger.debug('Start storing chunk metas for data keys: %s, '
                     'subtask id: %s', stored_keys, self.subtask.subtask_id)
        if set_chunk_metas:
            set_chunks_meta = asyncio.create_task(
                self._meta_api.set_chunk_meta.batch(*set_chunk_metas))
            try:
                await set_chunks_meta
                logger.debug('Finish store chunk metas for data keys: %s, '
                             'subtask id: %s', stored_keys, self.subtask.subtask_id)
            except asyncio.CancelledError:
                logger.debug('Cancelling store chunk metas for data keys: %s, '
                             'subtask id: %s', stored_keys, self.subtask.subtask_id)
                set_chunks_meta.cancel()

                # remote stored data
                deletes = []
                for data_key in stored_keys:
                    deletes.append(self._storage_api.delete.delay(data_key))
                await self._storage_api.delete.batch(*deletes)

                self.result.status = SubtaskStatus.cancelled
                logger.debug('Cancelled store chunk metas for data keys: %s, '
                             'subtask id: %s', stored_keys, self.subtask.subtask_id)
                raise
        # set result data size
        self.result.data_size = result_data_size

    async def done(self):
        if self.result.status == SubtaskStatus.running:
            self.result.status = SubtaskStatus.succeeded
        self.result.progress = 1.0
        self.is_done.set()

    async def run(self):
        self.result.status = SubtaskStatus.running
        input_keys = None
        unpinned = False
        try:
            chunk_graph = optimize(self._chunk_graph, self._engines)
            self._chunk_key_to_data_keys = get_chunk_key_to_data_keys(chunk_graph)
            report_progress = asyncio.create_task(
                self.report_progress_periodically())

            # load inputs data
            input_keys = await self._load_input_data()
            try:
                # execute chunk graph
                await self._execute_graph(chunk_graph)
            finally:
                # unpin inputs data
                unpinned = True
                await self._unpin_data(input_keys)
            # store results data
            stored_keys, store_sizes, memory_sizes, data_key_to_object_id = await self._store_data(chunk_graph)
            # store meta
            await self._store_meta(chunk_graph, stored_keys, store_sizes, memory_sizes, data_key_to_object_id)
        except asyncio.CancelledError:
            self.result.status = SubtaskStatus.cancelled
            self.result.progress = 1.0
            raise
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            self.result.status = SubtaskStatus.errored
            self.result.progress = 1.0
            _, self.result.error, self.result.traceback = sys.exc_info()
            await self.done()
            raise
        finally:
            if input_keys is not None and not unpinned:
                await self._unpin_data(input_keys)

        await self.done()
        report_progress.cancel()
        try:
            await report_progress
        except asyncio.CancelledError:
            pass
        return self.result

    async def report_progress_periodically(self, interval=.5, eps=0.001):
        last_progress = self.result.progress
        while not self.result.status.is_done:
            size = self._actual_chunk_count
            progress = sum(self._op_progress.values()) / size
            assert progress <= 1
            self.result.progress = progress
            if abs(last_progress - progress) >= eps:
                # report progress
                if not self.result.status.is_done:
                    fut = self.notify_task_manager_result(
                        self._supervisor_address, self.result)
                    if fut:
                        await fut
            await asyncio.sleep(interval)
            last_progress = progress


class SubtaskProcessorActor(mo.Actor):
    _session_api: Optional[SessionAPI]
    _storage_api: Optional[StorageAPI]
    _meta_api: Optional[MetaAPI]
    _processor: Optional[SubtaskProcessor]
    _last_processor: Optional[SubtaskProcessor]
    _running_aio_task: Optional[asyncio.Task]

    def __init__(self,
                 session_id: str,
                 band: BandType,
                 supervisor_address: str,
                 worker_address: str,
                 subtask_processor_cls: Type[SubtaskProcessor]):
        self._session_id = session_id
        self._band = band
        self._supervisor_address = supervisor_address
        self._worker_address = worker_address
        self._subtask_processor_cls = subtask_processor_cls

        # current processor
        self._processor = None
        self._last_processor = None
        self._running_aio_task = None

        self._session_api = None
        self._storage_api = None
        self._meta_api = None

    @classmethod
    def gen_uid(cls, session_id: str):
        return f'{session_id}_subtask_processor'

    async def __post_create__(self):
        coros = [
            SessionAPI.create(self._supervisor_address),
            StorageAPI.create(self._session_id, self.address, self._band[1]),
            MetaAPI.create(self._session_id, self._supervisor_address)]
        coros = [asyncio.ensure_future(coro) for coro in coros]
        await asyncio.gather(*coros)
        self._session_api, self._storage_api, self._meta_api = \
            [coro.result() for coro in coros]

    async def _init_context(self, session_id: str):
        loop = asyncio.get_running_loop()
        context = ThreadedServiceContext(
            session_id,
            self._supervisor_address, self._worker_address, self.address,
            loop, band=self._band)
        await context.init()
        set_context(context)

    async def run(self, subtask: Subtask):
        logger.debug('Start to run subtask: %s', subtask.subtask_id)

        assert subtask.session_id == self._session_id

        # init context
        await self._init_context(self._session_id)
        processor = self._subtask_processor_cls(
            subtask, self._session_api, self._storage_api, self._meta_api,
            self._band, self._supervisor_address)
        self._processor = self._last_processor = processor
        self._running_aio_task = asyncio.create_task(processor.run())
        try:
            result = yield self._running_aio_task
            raise mo.Return(result)
        finally:
            self._processor = self._running_aio_task = None

    async def wait(self):
        return self._processor.is_done.wait()

    async def result(self):
        return self._last_processor.result

    async def cancel(self):
        logger.debug('Cancelling subtask: %s', self._processor.subtask_id)

        aio_task = self._running_aio_task
        aio_task.cancel()

        async def waiter():
            try:
                await aio_task
            except asyncio.CancelledError:
                pass

        # return asyncio task to not block current actor
        return waiter()

    def get_running_subtask_id(self):
        return self._processor.subtask_id

    def set_running_op_progress(self, op_key: str, progress: float):
        self._processor.set_op_progress(op_key, progress)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Tuple, Union, Type

from ...utils import to_binary
from ..api import Actor
from ..core import ActorRef
from ..context import BaseActorContext
from ..debug import debug_async_timeout, detect_cycle_send
from ..errors import CannotCancelTask
from ..utils import create_actor_ref
from .allocate_strategy import AllocateStrategy, AddressSpecified
from .core import ActorCaller
from .message import DEFAULT_PROTOCOL, new_message_id, _MessageBase, \
    ResultMessage, ErrorMessage, CreateActorMessage, HasActorMessage, \
    DestroyActorMessage, ActorRefMessage, SendMessage, CancelMessage, \
    ControlMessage, ControlMessageType
from .router import Router


class MarsActorContext(BaseActorContext):
    __slots__ = '_address', '_caller'

    support_allocate_strategy = True

    def __init__(self, address: str = None):
        self._address = address
        self._caller = ActorCaller()

    def __del__(self):
        self._caller.cancel_tasks()

    async def _call(self,
                    address: str,
                    message: _MessageBase,
                    wait: bool = True) \
            -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        return await self._caller.call(Router.get_instance_or_empty(),
                                       address, message, wait=wait)

    @staticmethod
    def _process_result_message(message: Union[ResultMessage, ErrorMessage]):
        if isinstance(message, ResultMessage):
            return message.result
        else:
            raise message.error.with_traceback(message.traceback)

    async def _wait(self,
                    future: asyncio.Future,
                    address: str,
                    message: _MessageBase):
        try:
            await asyncio.wait([future])
        except asyncio.CancelledError:
            try:
                await self.cancel(address, message.message_id)
            except CannotCancelTask:
                # cancel failed, already finished
                raise asyncio.CancelledError
        return await future

    async def create_actor(self, actor_cls: Type[Actor], *args, uid=None,
                           address: str = None, **kwargs) -> ActorRef:
        router = Router.get_instance_or_empty()
        address = address or self._address or router.external_address
        allocate_strategy = kwargs.get('allocate_strategy', None)
        if isinstance(allocate_strategy, AllocateStrategy):
            allocate_strategy = kwargs.pop('allocate_strategy')
        else:
            allocate_strategy = AddressSpecified(address)
        create_actor_message = CreateActorMessage(
            new_message_id(), actor_cls, to_binary(uid),
            args, kwargs, allocate_strategy,
            protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(
            address, create_actor_message, wait=False)
        result = await self._wait(future, address, create_actor_message)
        return self._process_result_message(result)

    async def has_actor(self, actor_ref: ActorRef) -> bool:
        message = HasActorMessage(
            new_message_id(), actor_ref,
            protocol=DEFAULT_PROTOCOL)
        future = await self._call(actor_ref.address, message, wait=False)
        result = await self._wait(future, actor_ref.address, message)
        return self._process_result_message(result)

    async def destroy_actor(self, actor_ref: ActorRef):
        message = DestroyActorMessage(
            new_message_id(), actor_ref,
            protocol=DEFAULT_PROTOCOL)
        future = await self._call(actor_ref.address, message, wait=False)
        result = await self._wait(future, actor_ref.address, message)
        return self._process_result_message(result)

    async def kill_actor(self, actor_ref: ActorRef, force: bool = True):
        # get main_pool_address
        control_message = ControlMessage(
            new_message_id(), actor_ref.address,
            ControlMessageType.get_config,
            'main_pool_address',
            protocol=DEFAULT_PROTOCOL)
        main_address = self._process_result_message(
            await self._call(actor_ref.address, control_message))
        real_actor_ref = await self.actor_ref(actor_ref)
        if real_actor_ref.address == main_address:
            raise ValueError('Cannot kill actor on main pool')
        stop_message = ControlMessage(
            new_message_id(), real_actor_ref.address,
            ControlMessageType.stop,
            # default timeout (3 secs) and force
            (3., force), protocol=DEFAULT_PROTOCOL)
        # stop server
        result = await self._call(main_address, stop_message)
        return self._process_result_message(result)

    async def actor_ref(self, *args, **kwargs):
        actor_ref = create_actor_ref(*args, **kwargs)
        message = ActorRefMessage(
            new_message_id(), actor_ref,
            protocol=DEFAULT_PROTOCOL)
        future = await self._call(actor_ref.address, message, wait=False)
        result = await self._wait(future, actor_ref.address, message)
        return self._process_result_message(result)

    async def send(self,
                   actor_ref: ActorRef,
                   message: Tuple,
                   wait_response: bool = True):
        message = SendMessage(
            new_message_id(), actor_ref,
            message, protocol=DEFAULT_PROTOCOL)

        with debug_async_timeout('actor_call_timeout', 'Calling %r on %s at %s timed out',
                                 message.content, actor_ref.uid, actor_ref.address):
            detect_cycle_send(message, wait_response)
            future = await self._call(actor_ref.address, message, wait=False)
            if wait_response:
                result = await self._wait(future, actor_ref.address, message)
                return self._process_result_message(result)
            else:
                return future

    async def cancel(self,
                     address: str,
                     cancel_message_id: bytes):
        message = CancelMessage(
            new_message_id(), address, cancel_message_id,
            protocol=DEFAULT_PROTOCOL)
        result = await self._call(address, message)
        return self._process_result_message(result)

    async def wait_actor_pool_recovered(self, address: str,
                                        main_address: str = None):
        if main_address is None:
            # get main_pool_address
            control_message = ControlMessage(
                new_message_id(), address,
                ControlMessageType.get_config,
                'main_pool_address',
                protocol=DEFAULT_PROTOCOL)
            main_address = self._process_result_message(
                await self._call(address, control_message))

        # if address is main pool, it is never recovered
        if address == main_address:
            return

        control_message = ControlMessage(
            new_message_id(), address,
            ControlMessageType.wait_pool_recovered,
            None,
            protocol=DEFAULT_PROTOCOL
        )
        self._process_result_message(
            await self._call(main_address, control_message))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

from ....core import ChunkGraph, ChunkType
from ....core.operand import VirtualOperand, Fetch
from ....utils import build_fuse_chunk


class Fusion:
    """
    Fuse chunks in a chunk graph into a single node.
    """

    def __init__(self, graph: ChunkGraph):
        self._graph = graph

    def _fuse_nodes(self, nodes_list: List[List[ChunkType]]) -> \
            Tuple[List[List[ChunkType]], List[ChunkType]]:
        fused_nodes = []

        for nodes in nodes_list:
            head_node = nodes[0]
            tail_node = nodes[-1]
            fuse_chunk = build_fuse_chunk(
                nodes, tail_node.op.get_fuse_op_cls(tail_node), None, None).data
            self._graph.add_node(fuse_chunk)
            try:
                result_index = self._graph.results.index(tail_node)
                self._graph.results[result_index] = fuse_chunk
            except ValueError:
                pass
            for node in self._graph.iter_successors(tail_node):
                self._graph.add_edge(fuse_chunk, node)
                # replace inputs
                node_inputs = node.inputs
                new_node_inputs = []
                for inp in node_inputs:
                    if inp is tail_node:
                        new_node_inputs.append(fuse_chunk)
                    else:
                        new_node_inputs.append(inp)
                node.inputs = new_node_inputs
            for node in self._graph.iter_predecessors(head_node):
                self._graph.add_edge(node, fuse_chunk)
            # TODO:judge compose is independent?
            for node in nodes:
                self._graph.remove_node(node)
            fused_nodes.append(fuse_chunk)

        return nodes_list, fused_nodes

    def fuse(self) -> Tuple[List[List[ChunkType]], List[ChunkType]]:
        fuses = []
        explored = set()
        # for those chunk in result sets, we should not do any fuse
        result_chunk_set = set(self._graph.result_chunks)

        for v in self._graph.topological_iter():
            if v in explored or v in result_chunk_set:
                continue
            if self._graph.count_successors(v) != 1:
                continue
            if len(v.op.outputs) != 1:  # pragma: no cover
                continue
            if v.op.gpu != self._graph.successors(v)[0].op.gpu:
                continue
            if isinstance(v.op, (VirtualOperand, Fetch)):  # pragma: no cover
                # cannot fuse virtual operand or fetch
                continue
            if v.op.scheduling_hint is not None and \
                    not v.op.scheduling_hint.can_be_fused():  # pragma: no cover
                # don't fuse operand that cannot be fused
                continue
            selected = [v]
            # add successors
            cur_node = self._graph.successors(v)[0]
            while self._graph.count_predecessors(cur_node) == 1 and \
                    not isinstance(cur_node.op, (VirtualOperand, Fetch)):
                selected.append(cur_node)
                if self._graph.count_successors(cur_node) != 1 or \
                        cur_node in result_chunk_set:
                    break
                else:
                    cur_node = self._graph.successors(cur_node)[0]
            if len(selected) > 1:
                explored.update(selected)
                fuses.append(list(selected))
        return self._fuse_nodes(fuses)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from types import TracebackType
from typing import Any, Type, Tuple, Dict, List

from ...lib.tblib import pickling_support
from ...serialization.core import Serializer, pickle, buffered
from ...utils import classproperty, dataslots, implements
from ..core import ActorRef


# make sure traceback can be pickled
pickling_support.install()


DEFAULT_PROTOCOL = 0


class MessageType(Enum):
    control = 0
    result = 1
    error = 2
    create_actor = 3
    destroy_actor = 4
    has_actor = 5
    actor_ref = 6
    send = 7
    tell = 8
    cancel = 9


class ControlMessageType(Enum):
    stop = 0
    restart = 1
    sync_config = 2
    get_config = 3
    wait_pool_recovered = 4


@dataslots
@dataclass
class MessageTraceItem:
    uid: str
    address: str
    method: str


class _MessageBase(ABC):
    __slots__ = 'protocol', 'message_id', 'message_trace'

    def __init__(self,
                 message_id: bytes,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        self.message_id = message_id
        if protocol is None:
            protocol = DEFAULT_PROTOCOL
        self.protocol = protocol
        # A message can be in the scope of other messages,
        # this is mainly used for detecting deadlocks,
        # e.g. Actor `A` sent a message(id: 1) to actor `B`,
        # in the processing of `B`, it sent back a message(id: 2) to `A`,
        # deadlock happens, because `A` is still waiting for reply from `B`.
        # In this case, the `scoped_message_ids` will be [1, 2],
        # `A` will find that id:1 already exists in inbox,
        # thus deadlock detected.
        self.message_trace = message_trace

    @classproperty
    @abstractmethod
    def message_type(self) -> MessageType:
        """
        Message type.

        Returns
        -------
        message_type: MessageType
            message type.
        """

    def __repr__(self):
        slots = _get_slots(self.__class__)
        values = ', '.join(['{}={!r}'.format(slot, getattr(self, slot)) for slot in slots])
        return '{}({})'.format(self.__class__.__name__, values)


class ControlMessage(_MessageBase):
    __slots__ = 'address', 'control_message_type', 'content'

    def __init__(self,
                 message_id: bytes,
                 address: str,
                 control_message_type: ControlMessageType,
                 content: Any,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.address = address
        self.control_message_type = control_message_type
        self.content = content

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.control


class ResultMessage(_MessageBase):
    __slots__ = 'result',

    def __init__(self,
                 message_id: bytes,
                 result: Any,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.result = result

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.result


class ErrorMessage(_MessageBase):
    __slots__ = 'error_type', 'error', 'traceback'

    def __init__(self,
                 message_id: bytes,
                 error_type: Type[BaseException],
                 error: BaseException,
                 traceback: TracebackType,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.error_type = error_type
        self.error = error
        self.traceback = traceback

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.error


class CreateActorMessage(_MessageBase):
    __slots__ = 'actor_cls', 'actor_id', 'args', 'kwargs', 'allocate_strategy'

    def __init__(self,
                 message_id: bytes,
                 actor_cls: Type,
                 actor_id: bytes,
                 args: Tuple,
                 kwargs: Dict,
                 allocate_strategy,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.actor_cls = actor_cls
        self.actor_id = actor_id
        self.args = args
        self.kwargs = kwargs
        self.allocate_strategy = allocate_strategy

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.create_actor


class DestroyActorMessage(_MessageBase):
    __slots__ = 'actor_ref', 'from_main'

    def __init__(self,
                 message_id: bytes,
                 actor_ref: ActorRef,
                 from_main: bool = False,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.actor_ref = actor_ref
        self.from_main = from_main

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.destroy_actor


class HasActorMessage(_MessageBase):
    __slots__ = 'actor_ref',

    def __init__(self,
                 message_id: bytes,
                 actor_ref: ActorRef,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.actor_ref = actor_ref

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.has_actor


class ActorRefMessage(_MessageBase):
    __slots__ = 'actor_ref',

    def __init__(self,
                 message_id: bytes,
                 actor_ref: ActorRef,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.actor_ref = actor_ref

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.actor_ref


class SendMessage(_MessageBase):
    __slots__ = 'actor_ref', 'content',

    def __init__(self,
                 message_id: bytes,
                 actor_ref: ActorRef,
                 content: Any,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.actor_ref = actor_ref
        self.content = content

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.send


class TellMessage(SendMessage):
    __slots__ = ()

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.tell


class CancelMessage(_MessageBase):
    __slots__ = 'address', 'cancel_message_id',

    def __init__(self,
                 message_id: bytes,
                 address: str,
                 cancel_message_id: bytes,
                 protocol: int = None,
                 message_trace: List[MessageTraceItem] = None):
        super().__init__(message_id,
                         protocol=protocol,
                         message_trace=message_trace)
        self.address = address
        self.cancel_message_id = cancel_message_id

    @classproperty
    @implements(_MessageBase.message_type)
    def message_type(self) -> MessageType:
        return MessageType.cancel


class DeserializeMessageFailed(Exception):
    def __init__(self, message_id):
        self.message_id = message_id

    def __str__(self):
        return f'Deserialize {self.message_id} failed'


class MessageSerializer(Serializer):
    serializer_name = 'actor_message'

    @buffered
    def serialize(self, obj: _MessageBase, context: Dict):
        assert obj.protocol == 0, 'only support protocol 0 for now'

        message_class = type(obj)
        to_serialize = [getattr(obj, slot) for slot in _get_slots(message_class)]
        header, buffers = yield to_serialize
        new_header = {
            'message_class': message_class,
            'message_id': obj.message_id,
            'protocol': obj.protocol,
            'attributes_header': header
        }
        return new_header, buffers

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        protocol = header['protocol']
        assert protocol == 0, 'only support protocol 0 for now'
        message_id = header['message_id']
        message_class = header['message_class']
        try:
            serialized = yield header['attributes_header'], buffers
            message = object.__new__(message_class)
            for slot, val in zip(_get_slots(message_class), serialized):
                setattr(message, slot, val)
            return message
        except pickle.UnpicklingError as e:  # pragma: no cover
            raise DeserializeMessageFailed(message_id) from e


# register message serializer
MessageSerializer.register(_MessageBase)


@lru_cache(20)
def _get_slots(message_cls: Type[_MessageBase]):
    slots = []
    for tp in message_cls.__mro__:
        if issubclass(tp, _MessageBase):
            slots.extend(tp.__slots__)
    return slots


def new_message_id():
    return os.urandom(32)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .core import Serializable, SerializableMeta
from .field import AnyField, IdentityField, BoolField, \
    Int8Field, Int16Field, Int32Field, Int64Field, \
    UInt8Field, UInt16Field, UInt32Field, UInt64Field, \
    Float16Field, Float32Field, Float64Field, Complex64Field, Complex128Field, \
    StringField, BytesField, KeyField, NDArrayField, \
    Datetime64Field, Timedelta64Field, DataTypeField, \
    IndexField, SeriesField, DataFrameField, IntervalArrayField, \
    SliceField, FunctionField, NamedTupleField, TZInfoField, \
    ListField, TupleField, DictField, ReferenceField, OneOfField
from .field_type import FieldTypes



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField, BoolField, ListField, Int64Field
from ..utils import parse_index, lazy_import
from .core import IncrementalIndexDatasource, IncrementalIndexDataSourceMixin

ray = lazy_import('ray')
# Ray Datasets is available in early preview at ray.data with Ray 1.6+
# (and ray.experimental.data in Ray 1.5)
ray_dataset = lazy_import('ray.data')
ray_exp_dataset = lazy_import('ray.experimental.data')
real_ray_dataset = ray_dataset or ray_exp_dataset


class DataFrameReadRayDataset(IncrementalIndexDatasource,
                              IncrementalIndexDataSourceMixin):
    _op_type_ = OperandDef.READ_RAYDATASET

    _refs = AnyField('refs')
    _columns = ListField('columns')
    _incremental_index = BoolField('incremental_index')
    _nrows = Int64Field('nrows')

    def __init__(self, refs=None, columns=None,
                 incremental_index=None, nrows=None,
                 **kw):
        super().__init__(_refs=refs, _columns=columns,
                         _incremental_index=incremental_index,
                         _nrows=nrows,
                         **kw)

    @property
    def refs(self):
        return self._refs

    @property
    def columns(self):
        return self._columns

    @property
    def incremental_index(self):
        return self._incremental_index

    @classmethod
    def _tile_partitioned(cls, op: 'DataFrameReadRayDataset'):
        out_df = op.outputs[0]
        shape = (np.nan, out_df.shape[1])
        dtypes = out_df.dtypes
        dataset = op.refs

        chunk_index = 0
        out_chunks = []
        for object_ref in dataset:
            chunk_op = op.copy().reset_key()
            chunk_op._refs = [object_ref]
            new_chunk = chunk_op.new_chunk(
                None, shape=shape, index=(chunk_index, 0),
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
                dtypes=dtypes)
            out_chunks.append(new_chunk)
            chunk_index += 1

        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(None, out_df.shape, dtypes=dtypes,
                                     index_value=out_df.index_value,
                                     columns_value=out_df.columns_value,
                                     chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def _tile(cls, op):
        return cls._tile_partitioned(op)

    @classmethod
    def execute(cls, ctx, op: 'DataFrameReadRayDataset'):
        out = op.outputs[0]
        ref = op.refs[0]

        df = ray.get(ref)
        ctx[out.key] = df

    def __call__(self, index_value=None, columns_value=None, dtypes=None):
        shape = (np.nan, len(dtypes))
        return self.new_dataframe(None, shape, dtypes=dtypes, index_value=index_value,
                                  columns_value=columns_value)


def read_raydataset(ds, columns=None,
                    incremental_index=False,
                    **kwargs):
    assert isinstance(ds, real_ray_dataset.Dataset)
    refs = ds.to_pandas()
    dtypes = ds.schema().empty_table().to_pandas().dtypes
    index_value = parse_index(pd.RangeIndex(-1))
    columns_value = parse_index(dtypes.index, store_data=True)

    op = DataFrameReadRayDataset(refs=refs, columns=columns,
                                 incremental_index=incremental_index)
    return op(index_value=index_value, columns_value=columns_value,
              dtypes=dtypes)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
from typing import List, Optional, Union

from ....core import TileableGraph, Tileable
from ....lib.aio import alru_cache
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import TaskResult, TaskStatus
from .core import AbstractTaskAPI


def _json_serial_task_result(result: Optional[TaskResult]):
    if result is None:
        return {}
    return {
        'task_id': result.task_id,
        'session_id': result.session_id,
        'stage_id': result.stage_id,
        'start_time': result.start_time,
        'end_time': result.end_time,
        'progress': result.progress,
        'status': result.status.value,
        'error': base64.b64encode(serialize_serializable(result.error)).decode()
        if result.error is not None else None,
        'traceback': base64.b64encode(serialize_serializable(result.traceback)).decode()
        if result.traceback is not None else None,
    }


def _json_deserial_task_result(d: dict) -> Optional[TaskResult]:
    if not d:
        return None
    if d['error'] is not None:
        d['error'] = deserialize_serializable(base64.b64decode(d['error']))
    if d['traceback'] is not None:
        d['traceback'] = deserialize_serializable(base64.b64decode(d['traceback']))
    d['status'] = TaskStatus(d['status'])
    return TaskResult(**d)


class TaskWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = '/api/session/(?P<session_id>[^/]+)/task'

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI
        return await ClusterAPI.create(self._supervisor_addr)

    @alru_cache(cache_exceptions=False)
    async def _get_oscar_task_api(self, session_id: str):
        from .oscar import TaskAPI
        cluster_api = await self._get_cluster_api()
        [address] = await cluster_api.get_supervisors_by_keys([session_id])
        return await TaskAPI.create(session_id, address)

    @web_api('', method='post')
    async def submit_tileable_graph(self, session_id: str):
        body_args = deserialize_serializable(self.request.body) \
            if self.request.body else None

        task_name = body_args.get('task_name', None) or None
        fuse_enabled = body_args.get('fuse')

        graph = body_args['graph']
        extra_config = body_args.get('extra_config', None)

        oscar_api = await self._get_oscar_task_api(session_id)
        task_id = await oscar_api.submit_tileable_graph(
            graph, task_name=task_name, fuse_enabled=fuse_enabled,
            extra_config=extra_config)
        self.write(task_id)

    @web_api('', method='get')
    async def get_task_results(self, session_id: str):
        progress = bool(int(self.get_argument('progress', '0')))
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_task_results(progress=progress)
        self.write(json.dumps({'tasks': [_json_serial_task_result(r) for r in res]}))

    @web_api('(?P<task_id>[^/]+)', method='get', arg_filter={'action': 'fetch_tileables'})
    async def get_fetch_tileables(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_fetch_tileables(task_id)
        self.write(serialize_serializable(res))

    @web_api('(?P<task_id>[^/]+)', method='get')
    async def get_task_result(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_task_result(task_id)
        self.write(json.dumps(_json_serial_task_result(res)))

    @web_api('(?P<task_id>[^/]+)/tileable_graph', method='get',
             arg_filter={'action': 'get_tileable_graph_as_json'})
    async def get_tileable_graph_as_json(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_tileable_graph_as_json(task_id)
        self.write(json.dumps(res))

    @web_api('(?P<task_id>[^/]+)/tileable_detail', method='get')
    async def get_tileable_details(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_tileable_details(task_id)
        self.write(json.dumps(res))

    @web_api('(?P<task_id>[^/]+)/(?P<tileable_id>[^/]+)/subtask', method='get')
    async def get_tileable_subtasks(self, session_id: str, task_id: str, tileable_id: str):
        with_input_output = (self.get_argument('with_input_output', 'false') == 'true')
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_tileable_subtasks(
            task_id, tileable_id, with_input_output)
        self.write(json.dumps(res))

    @web_api('(?P<task_id>[^/]+)', method='get', arg_filter={'action': 'progress'})
    async def get_task_progress(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_task_progress(task_id)
        self.write(str(res))

    @web_api('', method='get', arg_filter={'action': 'last_idle_time'})
    async def get_last_idle_time(self, session_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.get_last_idle_time()
        if res:
            self.write(str(res))

    @web_api('(?P<task_id>[^/]+)', method='get', arg_filter={'action': 'wait'})
    async def wait_task(self, session_id: str, task_id: str):
        timeout = self.get_argument('timeout', None) or None
        timeout = float(timeout) if timeout is not None else None
        oscar_api = await self._get_oscar_task_api(session_id)
        res = await oscar_api.wait_task(task_id, timeout)
        self.write(json.dumps(_json_serial_task_result(res)))

    @web_api('(?P<task_id>[^/]+)', method='delete')
    async def cancel_task(self, session_id: str, task_id: str):
        oscar_api = await self._get_oscar_task_api(session_id)
        await oscar_api.cancel_task(task_id)


web_handlers = {
    TaskWebAPIHandler.get_root_pattern(): TaskWebAPIHandler
}


class WebTaskAPI(AbstractTaskAPI, MarsWebAPIClientMixin):
    def __init__(self, session_id: str, address: str):
        self._session_id = session_id
        self._address = address.rstrip('/')

    async def get_task_results(self, progress: bool = False) -> List[TaskResult]:
        path = f'{self._address}/api/session/{self._session_id}/task'
        params = {'progress': int(progress)}
        res = await self._request_url('GET', path, params=params)
        return [_json_deserial_task_result(d)
                for d in json.loads(res.body.decode())['tasks']]

    async def submit_tileable_graph(self,
                                    graph: TileableGraph,
                                    task_name: str = None,
                                    fuse_enabled: bool = True,
                                    extra_config: dict = None) -> str:
        path = f'{self._address}/api/session/{self._session_id}/task'
        extra_config_ser = serialize_serializable(extra_config) \
            if extra_config else None
        body = serialize_serializable({
            'task_name': task_name if task_name else '',
            'fuse': fuse_enabled,
            'graph': graph,
            'extra_config': extra_config_ser,
        })
        res = await self._request_url(
            path=path, method='POST',
            headers={'Content-Type': 'application/octet-stream'},
            data=body
        )
        return res.body.decode()

    async def get_fetch_tileables(self, task_id: str) -> List[Tileable]:
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}' \
               f'?action=fetch_tileables'
        res = await self._request_url('GET', path)
        return deserialize_serializable(res.body)

    async def get_task_result(self, task_id: str) -> TaskResult:
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}'
        res = await self._request_url('GET', path)
        return _json_deserial_task_result(json.loads(res.body.decode()))

    async def get_task_progress(self,
                                task_id: str) -> float:
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}'
        params = dict(action='progress')
        res = await self._request_url('GET', path, params=params)
        return float(res.body.decode())

    async def get_last_idle_time(self) -> Union[float, None]:
        path = f'{self._address}/api/session/{self._session_id}/task'
        params = dict(action='last_idle_time')
        res = await self._request_url('GET', path, params=params)
        content = res.body.decode()
        return float(content) if content else None

    async def wait_task(self, task_id: str, timeout: float = None):
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}'
        params = {'action': 'wait', 'timeout': str(timeout or '')}
        res = await self._request_url('GET', path, params=params,
                                      request_timeout=timeout or 0)
        return _json_deserial_task_result(json.loads(res.body.decode()))

    async def cancel_task(self, task_id: str):
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}'
        await self._request_url(path=path, method='DELETE')

    async def get_tileable_graph_as_json(self, task_id: str):
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}/tileable_graph'
        params = dict(action='get_tileable_graph_as_json')
        res = await self._request_url(path=path, params=params, method='GET')
        return json.loads(res.body.decode())

    async def get_tileable_details(self, task_id: str):
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}/tileable_detail'
        res = await self._request_url(path=path, method='GET')
        return json.loads(res.body.decode())

    async def get_tileable_subtasks(self,
                                    task_id: str,
                                    tileable_id: str,
                                    with_input_output: bool):

        with_input_output = 'true' if with_input_output else 'false'
        path = f'{self._address}/api/session/{self._session_id}/task/{task_id}/{tileable_id}/subtask'
        params = {
            'action': 'fetch_graph',
            'with_input_output': with_input_output,
        }
        res = await self._request_url(path=path, params=params, method='GET')
        return json.loads(res.body.decode())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from collections.abc import Iterable

import cloudpickle
import numpy as np
import pandas as pd
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy

from .version import parse as parse_version

_HAS_SQUEEZE = parse_version(pd.__version__) < parse_version('1.1.0')


class GroupByWrapper:
    def __init__(self, obj, groupby_obj=None, keys=None, axis=0, level=None, grouper=None,
                 exclusions=None, selection=None, as_index=True, sort=True,
                 group_keys=True, squeeze=False, observed=False, mutated=False,
                 grouper_cache=None):

        def fill_value(v, key):
            return v if v is not None or groupby_obj is None else getattr(groupby_obj, key)

        self.obj = obj
        self.keys = fill_value(keys, 'keys')
        self.axis = fill_value(axis, 'axis')
        self.level = fill_value(level, 'level')
        self.exclusions = fill_value(exclusions, 'exclusions')
        self.selection = selection
        self.as_index = fill_value(as_index, 'as_index')
        self.sort = fill_value(sort, 'sort')
        self.group_keys = fill_value(group_keys, 'group_keys')
        self.squeeze = fill_value(squeeze, 'squeeze')
        self.observed = fill_value(observed, 'observed')
        self.mutated = fill_value(mutated, 'mutated')

        if groupby_obj is None:
            groupby_kw = dict(keys=keys, axis=axis, level=level, grouper=grouper,
                              exclusions=exclusions, as_index=as_index, group_keys=group_keys,
                              squeeze=squeeze, observed=observed, mutated=mutated)
            if not _HAS_SQUEEZE:  # pragma: no branch
                groupby_kw.pop('squeeze')

            if obj.ndim == 2:
                self.groupby_obj = DataFrameGroupBy(obj, **groupby_kw)
            else:
                self.groupby_obj = SeriesGroupBy(obj, **groupby_kw)
        else:
            self.groupby_obj = groupby_obj

        if grouper_cache:
            self.groupby_obj.grouper._cache = grouper_cache
        if selection:
            self.groupby_obj = self.groupby_obj[selection]

        self.is_frame = isinstance(self.groupby_obj, DataFrameGroupBy)

    def __getitem__(self, item):
        return GroupByWrapper(
            self.obj, keys=self.keys, axis=self.axis, level=self.level,
            grouper=self.groupby_obj.grouper, exclusions=self.exclusions, selection=item,
            as_index=self.as_index, sort=self.sort, group_keys=self.group_keys,
            squeeze=self.squeeze, observed=self.observed, mutated=self.mutated)

    def __getattr__(self, item):
        if item.startswith('_'):  # pragma: no cover
            return object.__getattribute__(self, item)
        if item in getattr(self.obj, 'columns', ()):
            return self.__getitem__(item)
        return getattr(self.groupby_obj, item)

    def __iter__(self):
        return self.groupby_obj.__iter__()

    def __sizeof__(self):
        return sys.getsizeof(self.obj) \
            + sys.getsizeof(getattr(self.groupby_obj.grouper, '_cache', None))

    def __reduce__(self):
        return type(self).from_tuple, (self.to_tuple(pickle_function=True, truncate=True),)

    def __bool__(self):
        return bool(np.prod(self.shape))

    @property
    def empty(self):
        return self.obj.empty

    @property
    def shape(self):
        shape = list(self.groupby_obj.obj.shape)
        if self.is_frame and self.selection:
            shape[1] = len(self.selection)
        return tuple(shape)

    @property
    def _selected_obj(self):
        return getattr(self.groupby_obj, '_selected_obj')

    def to_tuple(self, truncate=False, pickle_function=False):
        if self.selection and truncate:
            if isinstance(self.selection, Iterable) and not isinstance(self.selection, str):
                item_list = list(self.selection)
            else:
                item_list = [self.selection]
            item_set = set(item_list)

            if isinstance(self.keys, list):
                sel_keys = self.keys
            elif self.keys in self.obj.columns:
                sel_keys = [self.keys]
            else:
                sel_keys = []

            all_items = item_list + [k for k in sel_keys or () if k not in item_set]
            if set(all_items) == set(self.obj.columns):
                obj = self.obj
            else:
                obj = self.obj[all_items]
        else:
            obj = self.obj

        if pickle_function and callable(self.keys):
            keys = cloudpickle.dumps(self.keys)
        else:
            keys = self.keys

        return obj, keys, self.axis, self.level, self.exclusions, self.selection, \
            self.as_index, self.sort, self.group_keys, self.squeeze, self.observed, \
            self.mutated, getattr(self.groupby_obj.grouper, '_cache', dict())

    @classmethod
    def from_tuple(cls, tp):
        obj, keys, axis, level, exclusions, selection, as_index, sort, group_keys, squeeze, \
            observed, mutated, grouper_cache = tp

        if isinstance(keys, (bytes, bytearray)):
            keys = cloudpickle.loads(keys)

        return cls(obj, keys=keys, axis=axis, level=level, exclusions=exclusions, selection=selection,
                   as_index=as_index, sort=sort, group_keys=group_keys, squeeze=squeeze, observed=observed,
                   mutated=mutated, grouper_cache=grouper_cache)


def wrapped_groupby(obj, by=None, axis=0, level=None, as_index=True, sort=True, group_keys=True,
                    squeeze=False, observed=False):
    groupby_kw = dict(by=by, axis=axis, level=level, as_index=as_index, sort=sort,
                      group_keys=group_keys, squeeze=squeeze, observed=observed)
    if not _HAS_SQUEEZE:  # pragma: no branch
        groupby_kw.pop('squeeze')

    groupby_obj = obj.groupby(**groupby_kw)
    return GroupByWrapper(obj, groupby_obj=groupby_obj)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Dict, Union

from ..errors import ServerClosed
from .communication import Client
from .message import _MessageBase, ResultMessage, ErrorMessage, \
    DeserializeMessageFailed
from .router import Router


result_message_type = Union[ResultMessage, ErrorMessage]


class ActorCaller:
    __slots__ = '_client_to_message_futures', '_clients'

    def __init__(self):
        self._client_to_message_futures: \
            Dict[Client, Dict[bytes, asyncio.Future]] = dict()
        self._clients: Dict[Client, asyncio.Task] = dict()

    async def get_client(self,
                         router: Router,
                         dest_address: str) -> Client:
        client = await router.get_client(dest_address,
                                         from_who=self)
        if client not in self._clients:
            self._clients[client] = asyncio.create_task(self._listen(client))
            self._client_to_message_futures[client] = dict()
        return client

    async def _listen(self, client: Client):
        while not client.closed:
            try:
                try:
                    message: _MessageBase = await client.recv()
                except (EOFError, ConnectionError, BrokenPipeError):
                    # remote server closed, close client and raise ServerClosed
                    try:
                        await client.close()
                    except (ConnectionError, BrokenPipeError):
                        # close failed, ignore it
                        pass
                    raise ServerClosed(f'Remote server {client.dest_address} closed') from None
                future = self._client_to_message_futures[client].pop(message.message_id)
                future.set_result(message)
            except DeserializeMessageFailed as e:  # pragma: no cover
                message_id = e.message_id
                future = self._client_to_message_futures[client].pop(message_id)
                future.set_exception(e)
            except Exception as e:  # noqa: E722  # pylint: disable=bare-except
                message_futures = self._client_to_message_futures.get(client)
                self._client_to_message_futures[client] = dict()
                for future in message_futures.values():
                    future.set_exception(e)

        message_futures = self._client_to_message_futures.get(client)
        self._client_to_message_futures[client] = dict()
        error = ServerClosed(f'Remote server {client.dest_address} closed')
        for future in message_futures.values():
            future.set_exception(error)

    async def call(self,
                   router: Router,
                   dest_address: str,
                   message: _MessageBase,
                   wait: bool = True) \
            -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        client = await self.get_client(router, dest_address)
        loop = asyncio.get_running_loop()
        wait_response = loop.create_future()
        self._client_to_message_futures[client][message.message_id] = wait_response

        try:
            await client.send(message)
        except ConnectionError:
            try:
                await client.close()
            except ConnectionError:
                # close failed, ignore it
                pass
            raise ServerClosed(f'Remote server {client.dest_address} closed')

        if not wait:
            return wait_response
        else:
            return await wait_response

    async def stop(self):
        try:
            await asyncio.gather(*[client.close() for client in self._clients])
        except (ConnectionError, ServerClosed):
            pass
        self.cancel_tasks()

    def cancel_tasks(self):
        # cancel listening for all clients
        _ = [task.cancel() for task in self._clients.values()]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Dict

from ..utils import lazy_import
from .core import Serializer, buffered

cupy = lazy_import('cupy', globals=globals())
cudf = lazy_import('cudf', globals=globals())


class CupySerializer(Serializer):
    serializer_name = 'cupy'

    @buffered
    def serialize(self, obj: Any, context: Dict):
        if not (obj.flags["C_CONTIGUOUS"] or obj.flags["F_CONTIGUOUS"]):
            obj = cupy.array(obj, copy=True)

        header = obj.__cuda_array_interface__.copy()
        header["strides"] = tuple(obj.strides)
        header["lengths"] = [obj.nbytes]
        buffer = cupy.ndarray(
            shape=(obj.nbytes,), dtype=cupy.dtype("u1"), memptr=obj.data, strides=(1,)
        )
        return header, [buffer]

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        return cupy.ndarray(
            shape=header["shape"],
            dtype=header["typestr"],
            memptr=cupy.asarray(buffers[0]).data,
            strides=header["strides"],
        )


class CudfSerializer(Serializer):
    serializer_name = 'cudf'

    def serialize(self, obj: Any, context: Dict):
        return obj.device_serialize()

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        from cudf.core.abc import Serializable
        return Serializable.device_deserialize(header, buffers)


if cupy is not None:
    CupySerializer.register('cupy.ndarray')
if cudf is not None:
    CudfSerializer.register('cudf.DataFrame')
    CudfSerializer.register('cudf.Series')
    CudfSerializer.register('cudf.Index')



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict, List, Union

from ....lib.aio import alru_cache
from ....utils import parse_readable_size
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import SessionInfo
from .core import AbstractSessionAPI


def _encode_size(size: Union[str, Dict[str, List[int]]]) -> str:
    if not isinstance(size, dict):
        return size
    else:
        return ','.join(f'{k}={v}' for k, v in size.items())


def _decode_size(encoded: str) -> Union[int, str, Dict[str, Union[int, List[int]]]]:
    if not encoded:
        return 0
    if ',' not in encoded and '=' not in encoded:
        try:
            return int(encoded)
        except ValueError:
            return int(parse_readable_size(encoded)[0])
    else:
        ret = dict()
        for kv in encoded.split(','):
            k, v = kv.split('=', 1)
            ret[k] = int(parse_readable_size(v)[0])
        return ret


class SessionWebAPIBaseHandler(MarsServiceWebAPIHandler):
    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI
        return await ClusterAPI.create(self._supervisor_addr)

    @alru_cache(cache_exceptions=False)
    async def _get_oscar_session_api(self):
        from .oscar import SessionAPI
        cluster_api = await self._get_cluster_api()
        [address] = await cluster_api.get_supervisors_by_keys(['Session'])
        return await SessionAPI.create(address)


class SessionWebAPIHandler(SessionWebAPIBaseHandler):
    @classmethod
    def get_root_pattern(cls):
        return '/api/session(?:/(?P<sub_path>[^/]*)$|$)'

    @web_api('(?P<session_id>[^/]+)', method='put')
    async def create_session(self, session_id: str):
        oscar_api = await self._get_oscar_session_api()
        addr = await oscar_api.create_session(session_id)
        self.write(addr)

    @web_api('(?P<session_id>[^/]+)', method='delete')
    async def delete_session(self, session_id: str):
        oscar_api = await self._get_oscar_session_api()
        await oscar_api.delete_session(session_id)

    @web_api('(?P<session_id>[^/]+)', method='get',
             arg_filter={'action': 'check_exist'})
    async def has_session(self, session_id: str):
        oscar_api = await self._get_oscar_session_api()
        res = await oscar_api.has_session(session_id)
        self.write('1' if res else '0')

    @web_api('(?P<session_id>[^/]*)', method='get',
             arg_filter={'action': 'get_last_idle_time'})
    async def get_last_idle_time(self, session_id: str):
        session_id = session_id or None
        oscar_api = await self._get_oscar_session_api()
        res = await oscar_api.get_last_idle_time(session_id)
        self.write(str(res) if res else '')

    @web_api('', method='get')
    async def get_sessions(self):
        oscar_api = await self._get_oscar_session_api()
        res = await oscar_api.get_sessions()
        self.write(json.dumps({'sessions': [
            {'session_id': info.session_id} for info in res
        ]}))


class SessionWebLogAPIHandler(SessionWebAPIBaseHandler):
    _root_pattern = '/api/session/(?P<session_id>[^/]+)/op/(?P<op_key>[^/]+)/log'

    @web_api('', method='get')
    async def fetch_tileable_op_logs(self,
                                     session_id: str,
                                     op_key: str):
        oscar_api = await self._get_oscar_session_api()
        offsets = _decode_size(self.get_argument('offsets', None))
        sizes = _decode_size(self.get_argument('sizes', None))
        log_result = await oscar_api.fetch_tileable_op_logs(
            session_id, op_key, offsets, sizes)
        self.write(json.dumps(log_result))


web_handlers = {
    SessionWebAPIHandler.get_root_pattern(): SessionWebAPIHandler,
    SessionWebLogAPIHandler.get_root_pattern(): SessionWebLogAPIHandler
}


class WebSessionAPI(AbstractSessionAPI, MarsWebAPIClientMixin):
    def __init__(self, address: str):
        self._address = address.rstrip('/')

    async def get_sessions(self) -> List[SessionInfo]:
        addr = f'{self._address}/api/session'
        res = await self._request_url('GET', addr)
        res_obj = json.loads(res.body.decode())
        return [SessionInfo(**kw) for kw in res_obj['sessions']]

    async def create_session(self, session_id: str) -> str:
        addr = f'{self._address}/api/session/{session_id}'
        res = await self._request_url(path=addr, method='PUT', data=b'')
        return res.body.decode()

    async def delete_session(self, session_id: str):
        addr = f'{self._address}/api/session/{session_id}'
        await self._request_url(path=addr, method='DELETE')

    async def has_session(self, session_id: str):
        addr = f'{self._address}/api/session/{session_id}'
        params = dict(action='check_exist')
        res = await self._request_url('GET', addr, params=params)
        return bool(int(res.body.decode()))

    async def get_last_idle_time(self, session_id: Union[str, None] = None) -> Union[float, None]:
        session_id = session_id or ''
        addr = f'{self._address}/api/session/{session_id}'
        params = dict(action='get_last_idle_time')
        res = await self._request_url('GET', addr, params=params)
        content = res.body.decode()
        return float(content) if content else None

    async def fetch_tileable_op_logs(self,
                                     session_id: str,
                                     tileable_op_key: str,
                                     chunk_op_key_to_offsets: Dict[str, List[int]],
                                     chunk_op_key_to_sizes: Dict[str, List[int]]) -> Dict:
        addr = f'{self._address}/api/session/{session_id}/op/{tileable_op_key}/log'
        params = dict(offsets=_encode_size(chunk_op_key_to_offsets),
                      sizes=_encode_size(chunk_op_key_to_sizes))
        res = await self._request_url('GET', addr, params=params)
        return json.loads(res.body.decode())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import socket
import sys
from typing import Dict, List

import numpy as np
import pandas as pd

try:
    import scipy
except ImportError:  # pragma: no cover
    scipy = None

from ... import resource as mars_resource
from ...config import options
from ...utils import git_info, lazy_import
from ...storage import StorageLevel
from .core import WorkerSlotInfo, QuotaInfo, DiskInfo, StorageInfo

cp = lazy_import('cupy', globals=globals(), rename='cp')
cudf = lazy_import('cudf', globals=globals())

logger = logging.getLogger(__name__)

_is_initial = True


def gather_node_env():
    from ...lib.mkl_interface import mkl_get_version
    from ...lib.nvutils import NVError
    from ..._version import __version__ as mars_version

    global _is_initial
    if _is_initial:
        _is_initial = False
        mars_resource.cpu_percent()

    mem_stats = mars_resource.virtual_memory()

    node_info = {
        'command_line': sys.argv,
        'platform': platform.platform(),
        'host_name': socket.gethostname(),
        'python_version': sys.version,
        'mars_version': mars_version,
        'cpu_total': mars_resource.cpu_count(),
        'memory_total': mem_stats.total,
        'options': options.to_dict(),
    }

    if 'MARS_K8S_POD_NAME' in os.environ:
        node_info['k8s_pod_name'] = os.environ['MARS_K8S_POD_NAME']
    if 'CONTAINER_ID' in os.environ:
        node_info['yarn_container_id'] = os.environ['CONTAINER_ID']

    try:
        cuda_info = mars_resource.cuda_info()
    except NVError:  # pragma: no cover
        logger.exception('NVError encountered, cannot gather CUDA devices.')
        cuda_info = None

    if cuda_info:
        node_info['cuda_info'] = {
            'driver': cuda_info.driver_version,
            'cuda': cuda_info.cuda_version,
            'products': list(cuda_info.products),
        }

    package_vers = {
        'numpy': np.__version__,
        'pandas': pd.__version__,
    }
    if hasattr(np, '__mkl_version__') and mkl_get_version:
        mkl_version = mkl_get_version()
        package_vers['mkl'] = f'{mkl_version.major}.{mkl_version.minor}.{mkl_version.update}'

    if scipy is not None:
        package_vers['scipy'] = scipy.__version__
    if cp is not None:
        package_vers['cupy'] = cp.__version__
    if cudf is not None:
        package_vers['cudf'] = cudf.__version__

    node_info['package_versions'] = package_vers

    git = git_info()
    if git:
        node_info['git_info'] = {
            'hash': git.commit_hash,
            'ref': git.commit_ref,
        }

    bands = node_info['bands'] = dict()

    cpu_band = {
        'resources': {
            'cpu': mars_resource.cpu_count(),
            'memory': mars_resource.virtual_memory().total,
        }
    }
    # todo numa can be supported by adding more bands
    bands['numa-0'] = cpu_band

    for idx, gpu_card_stat in enumerate(mars_resource.cuda_card_stats()):  # pragma: no cover
        bands[f'gpu-{idx}'] = {
            'resources': {
                'gpu': 1,
                'memory': gpu_card_stat.fb_mem_info.total,
            }
        }
    return node_info


def gather_node_resource(band_to_slots=None, use_gpu=True):
    # todo numa can be supported by adding more bands
    res = dict()
    mem_info = mars_resource.virtual_memory()
    num_cpu = mars_resource.cpu_count() if band_to_slots is None \
        else band_to_slots.get('numa-0', None)
    if num_cpu:  # pragma: no branch
        res['numa-0'] = {
            'cpu_avail': mars_resource.cpu_count() - mars_resource.cpu_percent() / 100.0,
            'cpu_total': num_cpu,
            'memory_avail': mem_info.available,
            'memory_total': mem_info.total,
        }

    if use_gpu:
        for idx, gpu_card_stat in enumerate(mars_resource.cuda_card_stats()):  # pragma: no cover
            num_gpu = 1 if band_to_slots is None else band_to_slots.get(f'gpu-{idx}')
            if not num_gpu:
                continue
            res[f'gpu-{idx}'] = {
                'gpu_avail': 1 - gpu_card_stat.gpu_usage,
                'gpu_total': num_gpu,
                'memory_avail': gpu_card_stat.fb_mem_info.available,
                'memory_total': gpu_card_stat.fb_mem_info.total,
            }
    return res


def gather_node_details(
    band_slot_infos: Dict[str, List[WorkerSlotInfo]] = None,
    band_quota_infos: Dict[str, QuotaInfo] = None,
    disk_infos: List[DiskInfo] = None,
    band_storage_infos: Dict[str, Dict[StorageLevel, StorageInfo]] = None
):
    disk_io_usage = mars_resource.disk_io_usage()
    net_io_usage = mars_resource.net_io_usage()
    res = {
        'disk': dict(zip(('reads', 'writes'), disk_io_usage)) if disk_io_usage else dict(),
        'network': dict(zip(('receives', 'sends'), net_io_usage)) if net_io_usage else dict(),
        'iowait': mars_resource.iowait(),
    }

    if disk_infos:
        part_dict = dict()
        for info in disk_infos:
            part_dev = mars_resource.get_path_device(info.path)
            if part_dev in part_dict:
                continue

            disk_usage_result = mars_resource.disk_usage(info.path)
            io_usage_result = mars_resource.disk_io_usage(info.path)
            part_dict[part_dev] = disk_info = {
                'size_limit': info.limit_size,
                'size_used': disk_usage_result.used,
                'size_total': disk_usage_result.total,
            }
            if io_usage_result is not None:
                disk_info.update({
                    'reads': io_usage_result.reads if io_usage_result else None,
                    'writes': io_usage_result.writes if io_usage_result else None,
                })
            if not sys.platform.startswith('win'):
                in_usage_result = os.statvfs(info.path)
                disk_info.update({
                    'inode_used': in_usage_result.f_files - in_usage_result.f_favail,
                    'inode_total': in_usage_result.f_files,
                })
        res['disk']['partitions'] = part_dict

    band_slot_infos = band_slot_infos or dict()
    res['slot'] = {
        band: [{
            'slot_id': slot_info.slot_id,
            'session_id': slot_info.session_id,
            'subtask_id': slot_info.subtask_id,
            'processor_usage': slot_info.processor_usage,
        } for slot_info in slot_infos]
        for band, slot_infos in band_slot_infos.items()
    }

    band_quota_infos = band_quota_infos or dict()
    res['quota'] = {
        band: {
            'quota_size': quota_info.quota_size,
            'allocated_size': quota_info.allocated_size,
            'hold_size': quota_info.hold_size,
        } for band, quota_info in band_quota_infos.items()
    }

    band_storage_infos = band_storage_infos or dict()
    res['storage'] = {
        band: {
            level.name.lower(): {
                'size_used': storage_info.used_size,
                'size_total': storage_info.total_size,
                'size_pinned': storage_info.pinned_size,
            }
            for level, storage_info in storage_infos.items()
        }
        for band, storage_infos in band_storage_infos.items()
    }
    return res



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from io import BytesIO
from typing import Any

import numpy as np

from ..utils import lazy_import
from .core import pickle, serialize, deserialize

cupy = lazy_import('cupy', globals=globals())
cudf = lazy_import('cudf', globals=globals())

DEFAULT_SERIALIZATION_VERSION = 0
BUFFER_SIZES_NAME = 'buf_sizes'


class AioSerializer:
    def __init__(self,
                 obj: Any,
                 compress=0):
        self._obj = obj
        self._compress = compress

    def _get_buffers(self):
        headers, buffers = serialize(self._obj)

        def _is_cuda_buffer(buf):  # pragma: no cover
            if cupy is not None and cudf is not None:
                from cudf.core import Buffer as CPBuffer
                from cupy import ndarray as cp_ndarray
            else:
                CPBuffer = cp_ndarray = None

            if CPBuffer is not None and isinstance(buf, CPBuffer):
                return True
            elif cp_ndarray is not None and isinstance(buf, cp_ndarray):
                return True
            else:
                return False

        is_cuda_buffers = [_is_cuda_buffer(buf) for buf in buffers]
        headers['is_cuda_buffers'] = np.array(is_cuda_buffers)

        # add buffer lengths into headers
        headers[BUFFER_SIZES_NAME] = [getattr(buf, 'nbytes', len(buf))
                                      for buf in buffers]
        header = pickle.dumps(headers)

        # gen header buffer
        header_bio = BytesIO()
        # write version first
        header_bio.write(struct.pack('B', DEFAULT_SERIALIZATION_VERSION))
        # write header length
        header_bio.write(struct.pack('<Q', len(header)))
        # write compression
        header_bio.write(struct.pack('<H', self._compress))
        # write header
        header_bio.write(header)

        out_buffers = list()
        out_buffers.append(header_bio.getbuffer())
        out_buffers.extend(buffers)

        return out_buffers

    async def run(self):
        return self._get_buffers()


class AioDeserializer:
    def __init__(self, file):
        self._file = file

    def _readexactly(self, n: int):
        # asyncio StreamReader may not guarantee to read n bytes
        # for it we need to call `readexactly` instead
        read = self._file.readexactly \
            if hasattr(self._file, 'readexactly') else self._file.read
        return read(n)

    async def _get_obj_header_bytes(self):
        try:
            header_bytes = bytes(await self._file.read(11))
        except ConnectionResetError:
            raise EOFError('Server may be closed')
        if len(header_bytes) == 0:
            raise EOFError('Received empty bytes')
        version = struct.unpack('B', header_bytes[:1])[0]
        # now we only have default version
        assert version == DEFAULT_SERIALIZATION_VERSION
        # header length
        header_length = struct.unpack('<Q', header_bytes[1:9])[0]
        # compress
        _ = struct.unpack('<H', header_bytes[9:])[0]
        return await self._readexactly(header_length)

    async def _get_obj(self):
        header = pickle.loads(await self._get_obj_header_bytes())
        # get buffer size
        buffer_sizes = header.pop(BUFFER_SIZES_NAME)
        # get buffers
        buffers = [await self._readexactly(size) for size in buffer_sizes]

        return deserialize(header, buffers)

    async def run(self):
        return await self._get_obj()

    async def get_size(self):
        # extract header
        header_bytes = await self._get_obj_header_bytes()
        header = pickle.loads(header_bytes)
        # get buffer size
        buffer_sizes = header.pop(BUFFER_SIZES_NAME)
        return 11 + len(header_bytes) + sum(buffer_sizes)

    async def get_header(self):
        return pickle.loads(await self._get_obj_header_bytes())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
import sys

from ... import oscar as mo
from ...lib.aio import alru_cache
from ..subtask import SubtaskResult, SubtaskStatus
from ..task import TaskAPI


@alru_cache
async def _get_task_api(actor: mo.Actor):
    return await TaskAPI.create(getattr(actor, '_session_id'), actor.address)


@contextlib.asynccontextmanager
async def redirect_subtask_errors(actor: mo.Actor, subtasks):
    try:
        yield
    except:  # noqa: E722  # pylint: disable=bare-except
        _, error, traceback = sys.exc_info()
        status = SubtaskStatus.cancelled if isinstance(error, asyncio.CancelledError) else SubtaskStatus.errored
        task_api = await _get_task_api(actor)
        coros = []
        for subtask in subtasks:
            if subtask is None:  # pragma: no cover
                continue
            coros.append(task_api.set_subtask_result(SubtaskResult(
                subtask_id=subtask.subtask_id,
                session_id=subtask.session_id,
                task_id=subtask.task_id,
                progress=1.0,
                status=status,
                error=error, traceback=traceback
            )))
        await asyncio.wait(coros)
        raise



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, Generator, Iterable, \
    List, Optional, Set, Tuple, Type, Union

from ....core import FUSE_CHUNK_TYPE, CHUNK_TYPE, TILEABLE_TYPE
from ....typing import EntityType, TileableType, ChunkType
from ....utils import copy_tileables, build_fetch
from ...entity.tileables import handler
from ...mode import enter_mode
from ..entity import TileableGraph, ChunkGraph
from .base import AbstractGraphBuilder


tile_gen_type = Generator[List[ChunkType], List[ChunkType], List[TileableType]]


class Tiler:
    _cur_chunk_graph: Optional[ChunkGraph]
    _tileable_handlers: Iterable[Tuple[TileableType, tile_gen_type]]

    def __init__(self,
                 tileable_graph: TileableGraph,
                 tile_context: Dict[TileableType, TileableType],
                 processed_chunks: Set[ChunkType],
                 add_nodes: Callable):
        self._tileable_graph = tileable_graph
        self._tile_context = tile_context
        self._processed_chunks = processed_chunks
        self._add_nodes = add_nodes
        self._cur_chunk_graph = None
        self._tileable_handlers = (
            (tileable, self._tile_handler(tileable))
            for tileable in tileable_graph.topological_iter())

    @staticmethod
    def _get_data(entity: EntityType):
        return entity.data if hasattr(entity, 'data') else entity

    def _tile_handler(self,
                      tileable: TileableType) -> tile_gen_type:
        from ....core.operand import Fetch

        tileable = self._get_data(tileable)

        if isinstance(tileable.op, Fetch) and not tileable.is_coarse():
            return [tileable]

        assert tileable.is_coarse()

        # copy tileable
        tiled_tileables = copy_tileables(
            tileable.op.outputs,
            inputs=[self._tile_context[inp] for inp in tileable.inputs],
            copy_key=True, copy_id=False)
        tiled_tileables = [self._get_data(t) for t in tiled_tileables]
        # start to tile
        tiled_tileables = yield from handler.tile(tiled_tileables)
        return tiled_tileables

    def _gen_tileable_handlers(self,
                               next_tileable_handlers: List[Tuple[TileableType, tile_gen_type]]):
        for tileable, tile_handler in self._tileable_handlers:
            if tileable in self._tile_context:
                continue
            if any(inp not in self._tile_context
                   for inp in self._tileable_graph.predecessors(tileable)):
                # predecessors not finished yet
                next_tileable_handlers.append((tileable, tile_handler))
                continue

            yield tileable, tile_handler

    def _tile(self,
              chunk_graph: ChunkGraph,
              tileable: TileableType,
              tile_handler: tile_gen_type,
              next_tileable_handlers: List[Tuple[TileableType, tile_gen_type]],
              to_update_tileables: List[TileableType],
              visited: Set[EntityType]):
        try:
            need_process = next(tile_handler)
            chunks = []
            if need_process is not None:
                for t in need_process:
                    if isinstance(t, CHUNK_TYPE):
                        chunks.append(self._get_data(t))
                    elif isinstance(t, TILEABLE_TYPE):
                        to_update_tileables.append(self._get_data(t))
            # not finished yet
            self._add_nodes(chunk_graph, chunks.copy(), visited)
            next_tileable_handlers.append((tileable, tile_handler))
            # add intermediate chunks into result chunks
            # to prevent them being pruned
            chunk_graph.result_chunks.extend(chunks)
        except StopIteration as e:
            # tile done
            tiled_tileables = e.value
            for out, tiled_tileable in zip(tileable.op.outputs, tiled_tileables):
                out = self._get_data(out)
                tiled_tileable = self._get_data(tiled_tileable)

                chunks = tiled_tileable.chunks
                if chunks is None:  # pragma: no cover
                    raise ValueError(f'tileable({out}) is still coarse '
                                     f'after tile')
                chunks = [self._get_data(c) for c in chunks]
                self._add_nodes(chunk_graph, chunks, visited)
                self._tile_context[out] = tiled_tileable

    def _gen_result_chunks(self,
                           chunk_graph: ChunkGraph,
                           next_tileable_handlers: List[Tuple[TileableType, tile_gen_type]]):
        result_chunks = chunk_graph.result_chunks
        tileable_graph = self._tileable_graph
        # generate result chunks
        result_chunk_set = set()
        if next_tileable_handlers:
            # add all chunks that have no successors to result chunks
            for chunk in chunk_graph:
                if chunk_graph.count_successors(chunk) == 0:
                    if chunk not in result_chunk_set:
                        result_chunks.append(chunk)
                        result_chunk_set.add(chunk)
            for tileable, _ in next_tileable_handlers:
                # tileable that tile not completed,
                # scan inputs to make sure their chunks in result
                for inp_tileable in tileable_graph.predecessors(tileable):
                    if inp_tileable in self._tile_context:
                        for chunk in self._tile_context[inp_tileable].chunks:
                            chunk = self._get_data(chunk)
                            if chunk in chunk_graph and \
                                    chunk not in result_chunk_set:
                                result_chunks.append(chunk)
                                result_chunk_set.add(chunk)
        for tileable in tileable_graph.result_tileables:
            if tileable in self._tile_context:
                for chunk in self._tile_context[tileable].chunks:
                    chunk = self._get_data(chunk)
                    if chunk in chunk_graph and \
                            chunk not in result_chunk_set:
                        result_chunks.append(chunk)
                        result_chunk_set.add(chunk)

    def _iter(self):
        chunk_graph = self._cur_chunk_graph

        to_update_tileables = []
        visited = set()

        if chunk_graph is not None:
            # last tiled chunks, add them to processed
            # so that fetch chunk can be generated
            processed_chunks = [
                c.chunk if isinstance(c, FUSE_CHUNK_TYPE) else c
                for c in chunk_graph.result_chunks]
            self._processed_chunks.update(processed_chunks)

        result_chunks = []
        chunk_graph = self._cur_chunk_graph = ChunkGraph(result_chunks)

        next_tileable_handlers = []
        # tile
        for tileable, tile_handler in \
                self._gen_tileable_handlers(next_tileable_handlers):
            self._tile(chunk_graph, tileable, tile_handler,
                       next_tileable_handlers, to_update_tileables, visited)
        self._tileable_handlers = next_tileable_handlers
        # gen result chunks
        self._gen_result_chunks(chunk_graph, next_tileable_handlers)

        return to_update_tileables

    def __iter__(self):
        while self._tileable_handlers:
            to_update_tileables = self._iter()
            yield self._cur_chunk_graph
            for t in to_update_tileables:
                t.refresh_params()


class ChunkGraphBuilder(AbstractGraphBuilder):
    _graph: TileableGraph

    def __init__(self,
                 graph: TileableGraph,
                 fuse_enabled: bool = True,
                 tile_context: Dict[TileableType, TileableType] = None,
                 tiler_cls: Union[Type[Tiler], Callable] = None):
        super().__init__(graph)
        self.fuse_enabled = fuse_enabled
        self.tile_context = dict() if tile_context is None else tile_context

        self._processed_chunks: Set[ChunkType] = set()
        self._chunk_to_fetch: Dict[ChunkType, ChunkType] = dict()

        tiler_cls = Tiler if tiler_cls is None else tiler_cls
        self.tiler = tiler_cls(self._graph, self.tile_context,
                               self._processed_chunks, self._add_nodes)

    def _select_inputs(self, inputs: List[ChunkType]):
        new_inputs = []
        for inp in inputs:
            # TODO: remove it when fuse chunk is deprecated
            if isinstance(inp, FUSE_CHUNK_TYPE):
                inp = inp.chunk
            if inp in self._processed_chunks:
                # gen fetch
                if inp not in self._chunk_to_fetch:
                    fetch_chunk = build_fetch(inp).data
                    self._chunk_to_fetch[inp] = fetch_chunk
                new_inputs.append(self._chunk_to_fetch[inp])
            else:
                new_inputs.append(inp)
        return new_inputs

    def _if_add_node(self, node: EntityType, visited: Set):
        return node not in visited and node not in self._processed_chunks

    def _build(self) -> Iterable[Union[TileableGraph, ChunkGraph]]:
        yield from self.tiler

    def build(self) -> Generator[Union[TileableGraph, ChunkGraph], None, None]:
        with enter_mode(build=True, kernel=True):
            yield from self._build()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List, Dict, Union

from ...services import start_services, stop_services, NodeRole
from ..utils import load_service_config_file


def load_config(filename=None):
    # use default config
    if not filename:  # pragma: no cover
        d = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(d, 'config.yml')
    return load_service_config_file(filename)


async def start_supervisor(address: str,
                           lookup_address: str = None,
                           modules: Union[List, str, None] = None,
                           config: Dict = None,
                           web: Union[str, bool] = 'auto'):
    if not config or isinstance(config, str):
        config = load_config(config)
    lookup_address = lookup_address or address
    backend = config['cluster'].get('backend', 'fixed')
    if backend == 'fixed' and config['cluster'].get('lookup_address') is None:
        config['cluster']['lookup_address'] = lookup_address
    if web:
        # try to append web to services
        config['services'].append('web')
    if modules:
        config['modules'] = modules
    try:
        await start_services(NodeRole.SUPERVISOR, config, address=address)
    except ImportError:
        if web == 'auto':
            config['services'] = [service for service in config['services']
                                  if service != 'web']
            await start_services(NodeRole.SUPERVISOR, config, address=address)
            return False
        else:  # pragma: no cover
            raise
    else:
        return bool(web)


async def stop_supervisor(address: str,
                          config: Dict = None):
    if not config or isinstance(config, str):
        config = load_config(config)
    await stop_services(NodeRole.SUPERVISOR, address=address, config=config)


async def start_worker(address: str,
                       lookup_address: str,
                       band_to_slots: Dict[str, int],
                       modules: Union[List, str, None] = None,
                       config: Dict = None,
                       mark_ready: bool = True):
    if not config or isinstance(config, str):
        config = load_config(config)
    backend = config['cluster'].get('backend', 'fixed')
    if backend == 'fixed' and config['cluster'].get('lookup_address') is None:
        config['cluster']['lookup_address'] = lookup_address
    if config['cluster'].get('resource') is None:
        config['cluster']['resource'] = band_to_slots
    if any(band_name.startswith('gpu-') for band_name in band_to_slots):  # pragma: no cover
        if 'cuda' not in config['storage']['backends']:
            config['storage']['backends'].append('cuda')
    if modules:
        config['modules'] = modules
    await start_services(NodeRole.WORKER, config, address=address,
                         mark_ready=mark_ready)


async def stop_worker(address: str,
                      config: Dict = None):
    if not config or isinstance(config, str):
        config = load_config(config)
    await stop_services(NodeRole.WORKER, address=address, config=config)
